#!/usr/bin/env python
"""
Numbers, names, and colormaps for the DKT and FreeSurfer labeling protocols
==============================================================================

The DKTprotocol class provides variables related to the
Desikan-Killiany-Tourville (DKT) human brain cortical labeling protocol.
For more information, see http://mindboggle.info/data/ and the article:

http://www.frontiersin.org/Brain_Imaging_Methods/10.3389/fnins.2012.00171/full
"101 labeled brain images and a consistent human cortical labeling protocol"
Arno Klein, Jason Tourville. Frontiers in Brain Imaging Methods. 6:171.
DOI: 10.3389/fnins.2012.00171

See http://mindboggle.info/faq/labels.html for additional information.

The return_numbers_names_colors() and extract_numbers_names_colors() functions
return numbers, names, and colors representing anatomical brain regions from
FreeSurfer's FreeSurferColorLUT.txt lookup table file.


Authors:
    - Jason Tourville, 2011-2012  (jtour@bu.edu)
    - Arno Klein, 2011-2015  (arno@mindboggle.info)  http://binarybottle.com

Copyright 2015,  Mindboggle team (http://mindboggle.info), Apache v2.0 License

"""


def extract_numbers_names_colors(FreeSurferColorLUT=''):
    """
    Extract lists of numbers, names, and colors representing anatomical brain
    regions from FreeSurfer's FreeSurferColorLUT.txt lookup table file.

    Parameters
    ----------
    FreeSurferColorLUT : string
        full path to FreeSurferColorLUT.txt file (else uses local Python file)

    Returns
    -------
    numbers : list of integers
        numbers representing anatomical labels from FreeSurferColorLUT.txt
    names : list of integers
        names for anatomical regions from FreeSurferColorLUT.txt
    colors : list of integers
        colors associated with anatomical labels from FreeSurferColorLUT.txt

    Examples
    --------
    >>> from mindboggle.mio.labels import extract_numbers_names_colors # doctest: +SKIP
    >>> ennc = extract_numbers_names_colors # doctest: +SKIP
    >>> en1,en2,ec = ennc('/Applications/freesurfer/FreeSurferColorLUT.txt') # doctest: +SKIP

    """
    import os
    from io import open

    from mindboggle.thirdparty.FreeSurferColorLUT import lut_text

    def is_number(s):
        try:
            int(s)
            return True
        except ValueError:
            return False

    # if os.environ['FREESURFER_HOME']:
    #     FreeSurferColorLUT = os.path.join(
    #              os.environ['FREESURFER_HOME'], 'FreeSurferColorLUT.txt')

    if FreeSurferColorLUT and os.path.exists(FreeSurferColorLUT):
        f = open(FreeSurferColorLUT, 'r')
        lines = f.readlines()
    else:
        lut = lut_text()
        lines = lut.split('\n')

    numbers = []
    names = []
    colors = []
    for line in lines:
        strings = line.split()
        if strings and is_number(strings[0]):
            numbers.append(int(strings[0]))
            names.append(strings[1])
            colors.append([int(strings[2]), int(strings[3]),
                           int(strings[4])])

    return numbers, names, colors


def return_numbers_names_colors():
    """
    Return lists of numbers, names, and colors representing anatomical brain
    regions derived from FreeSurfer's FreeSurferColorLUT.txt lookup file.

    Returns
    -------
    numbers : list of integers
        numbers representing anatomical labels from FreeSurferColorLUT.txt
    names : list of integers
        names for anatomical regions from FreeSurferColorLUT.txt
    colors : list of integers
        colors associated with anatomical labels from FreeSurferColorLUT.txt

    Examples
    --------
    >>> import numpy as np
    >>> from mindboggle.mio.labels import return_numbers_names_colors
    >>> numbers, names, colors = return_numbers_names_colors()
    >>> len(np.unique(numbers))
    1266
    >>> names[0:3]
    ['Unknown', 'Left-Cerebral-Exterior', 'Left-Cerebral-White-Matter']
    >>> colors[0:3]
    [[0, 0, 0], [70, 130, 180], [245, 245, 245]]

    """

    numbers = [0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
        13,
        14,
        15,
        16,
        17,
        18,
        19,
        20,
        21,
        22,
        23,
        24,
        25,
        26,
        27,
        28,
        29,
        30,
        31,
        32,
        33,
        34,
        35,
        36,
        37,
        38,
        39,
        40,
        41,
        42,
        43,
        44,
        45,
        46,
        47,
        48,
        49,
        50,
        51,
        52,
        53,
        54,
        55,
        56,
        57,
        58,
        59,
        60,
        61,
        62,
        63,
        64,
        65,
        66,
        67,
        68,
        69,
        70,
        71,
        72,
        73,
        74,
        77,
        78,
        79,
        80,
        81,
        82,
        83,
        84,
        85,
        192,
        86,
        87,
        88,
        96,
        97,
        98,
        100,
        101,
        102,
        103,
        104,
        105,
        106,
        107,
        108,
        109,
        110,
        111,
        112,
        113,
        114,
        115,
        116,
        117,
        118,
        119,
        120,
        121,
        122,
        123,
        124,
        125,
        126,
        127,
        128,
        129,
        130,
        131,
        132,
        133,
        134,
        135,
        136,
        137,
        138,
        139,
        140,
        142,
        143,
        144,
        145,
        146,
        147,
        148,
        149,
        150,
        151,
        152,
        153,
        154,
        155,
        156,
        157,
        158,
        159,
        160,
        161,
        162,
        163,
        164,
        165,
        166,
        167,
        168,
        169,
        176,
        170,
        171,
        172,
        173,
        174,
        175,
        180,
        181,
        193,
        194,
        195,
        196,
        197,
        198,
        199,
        200,
        201,
        202,
        203,
        204,
        205,
        206,
        207,
        208,
        209,
        210,
        211,
        212,
        213,
        214,
        215,
        216,
        217,
        218,
        219,
        220,
        221,
        222,
        223,
        224,
        225,
        226,
        250,
        251,
        252,
        253,
        254,
        255,
        256,
        331,
        332,
        333,
        334,
        335,
        336,
        337,
        338,
        339,
        340,
        341,
        342,
        343,
        344,
        345,
        346,
        347,
        348,
        349,
        350,
        351,
        352,
        353,
        354,
        355,
        356,
        357,
        358,
        359,
        400,
        401,
        402,
        403,
        404,
        405,
        406,
        407,
        408,
        409,
        410,
        411,
        412,
        413,
        414,
        415,
        416,
        417,
        418,
        419,
        420,
        421,
        422,
        423,
        424,
        425,
        426,
        427,
        428,
        429,
        430,
        431,
        432,
        433,
        434,
        435,
        436,
        437,
        438,
        439,
        498,
        499,
        500,
        501,
        502,
        503,
        504,
        505,
        506,
        507,
        508,
        550,
        551,
        552,
        553,
        554,
        555,
        556,
        557,
        558,
        600,
        601,
        602,
        603,
        604,
        605,
        606,
        607,
        608,
        609,
        610,
        611,
        612,
        613,
        614,
        615,
        616,
        617,
        618,
        619,
        620,
        621,
        622,
        623,
        624,
        625,
        626,
        627,
        628,
        640,
        641,
        642,
        643,
        644,
        645,
        646,
        647,
        648,
        649,
        650,
        651,
        652,
        653,
        654,
        655,
        656,
        657,
        658,
        659,
        660,
        661,
        662,
        663,
        664,
        665,
        666,
        667,
        668,
        669,
        670,
        671,
        672,
        673,
        674,
        675,
        676,
        677,
        678,
        679,
        701,
        702,
        703,
        999,
        1000,
        1001,
        1002,
        1003,
        1004,
        1005,
        1006,
        1007,
        1008,
        1009,
        1010,
        1011,
        1012,
        1013,
        1014,
        1015,
        1016,
        1017,
        1018,
        1019,
        1020,
        1021,
        1022,
        1023,
        1024,
        1025,
        1026,
        1027,
        1028,
        1029,
        1030,
        1031,
        1032,
        1033,
        1034,
        1035,
        2000,
        2001,
        2002,
        2003,
        2004,
        2005,
        2006,
        2007,
        2008,
        2009,
        2010,
        2011,
        2012,
        2013,
        2014,
        2015,
        2016,
        2017,
        2018,
        2019,
        2020,
        2021,
        2022,
        2023,
        2024,
        2025,
        2026,
        2027,
        2028,
        2029,
        2030,
        2031,
        2032,
        2033,
        2034,
        2035,
        3000,
        3001,
        3002,
        3003,
        3004,
        3005,
        3006,
        3007,
        3008,
        3009,
        3010,
        3011,
        3012,
        3013,
        3014,
        3015,
        3016,
        3017,
        3018,
        3019,
        3020,
        3021,
        3022,
        3023,
        3024,
        3025,
        3026,
        3027,
        3028,
        3029,
        3030,
        3031,
        3032,
        3033,
        3034,
        3035,
        4000,
        4001,
        4002,
        4003,
        4004,
        4005,
        4006,
        4007,
        4008,
        4009,
        4010,
        4011,
        4012,
        4013,
        4014,
        4015,
        4016,
        4017,
        4018,
        4019,
        4020,
        4021,
        4022,
        4023,
        4024,
        4025,
        4026,
        4027,
        4028,
        4029,
        4030,
        4031,
        4032,
        4033,
        4034,
        4035,
        1100,
        1101,
        1102,
        1103,
        1104,
        1200,
        1201,
        1202,
        1205,
        1206,
        1207,
        1210,
        1211,
        1212,
        1105,
        1106,
        1107,
        1108,
        1109,
        1110,
        1111,
        1112,
        1113,
        1114,
        1115,
        1116,
        1117,
        1118,
        1119,
        1120,
        1121,
        1122,
        1123,
        1124,
        1125,
        1126,
        1127,
        1128,
        1129,
        1130,
        1131,
        1132,
        1133,
        1134,
        1135,
        1136,
        1137,
        1138,
        1139,
        1140,
        1141,
        1142,
        1143,
        1144,
        1145,
        1146,
        1147,
        1148,
        1149,
        1150,
        1151,
        1152,
        1153,
        1154,
        1155,
        1156,
        1157,
        1158,
        1159,
        1160,
        1161,
        1162,
        1163,
        1164,
        1165,
        1166,
        1167,
        1168,
        1169,
        1170,
        1171,
        1172,
        1173,
        1174,
        1175,
        1176,
        1177,
        1178,
        1179,
        1180,
        1181,
        2100,
        2101,
        2102,
        2103,
        2104,
        2105,
        2106,
        2107,
        2108,
        2109,
        2110,
        2111,
        2112,
        2113,
        2114,
        2115,
        2116,
        2117,
        2118,
        2119,
        2120,
        2121,
        2122,
        2123,
        2124,
        2125,
        2126,
        2127,
        2128,
        2129,
        2130,
        2131,
        2132,
        2133,
        2134,
        2135,
        2136,
        2137,
        2138,
        2139,
        2140,
        2141,
        2142,
        2143,
        2144,
        2145,
        2146,
        2147,
        2148,
        2149,
        2150,
        2151,
        2152,
        2153,
        2154,
        2155,
        2156,
        2157,
        2158,
        2159,
        2160,
        2161,
        2162,
        2163,
        2164,
        2165,
        2166,
        2167,
        2168,
        2169,
        2170,
        2171,
        2172,
        2173,
        2174,
        2175,
        2176,
        2177,
        2178,
        2179,
        2180,
        2181,
        2200,
        2201,
        2202,
        2205,
        2206,
        2207,
        2210,
        2211,
        2212,
        3100,
        3101,
        3102,
        3103,
        3104,
        3105,
        3106,
        3107,
        3108,
        3109,
        3110,
        3111,
        3112,
        3113,
        3114,
        3115,
        3116,
        3117,
        3118,
        3119,
        3120,
        3121,
        3122,
        3123,
        3124,
        3125,
        3126,
        3127,
        3128,
        3129,
        3130,
        3131,
        3132,
        3133,
        3134,
        3135,
        3136,
        3137,
        3138,
        3139,
        3140,
        3141,
        3142,
        3143,
        3144,
        3145,
        3146,
        3147,
        3148,
        3149,
        3150,
        3151,
        3152,
        3153,
        3154,
        3155,
        3156,
        3157,
        3158,
        3159,
        3160,
        3161,
        3162,
        3163,
        3164,
        3165,
        3166,
        3167,
        3168,
        3169,
        3170,
        3171,
        3172,
        3173,
        3174,
        3175,
        3176,
        3177,
        3178,
        3179,
        3180,
        3181,
        4100,
        4101,
        4102,
        4103,
        4104,
        4105,
        4106,
        4107,
        4108,
        4109,
        4110,
        4111,
        4112,
        4113,
        4114,
        4115,
        4116,
        4117,
        4118,
        4119,
        4120,
        4121,
        4122,
        4123,
        4124,
        4125,
        4126,
        4127,
        4128,
        4129,
        4130,
        4131,
        4132,
        4133,
        4134,
        4135,
        4136,
        4137,
        4138,
        4139,
        4140,
        4141,
        4142,
        4143,
        4144,
        4145,
        4146,
        4147,
        4148,
        4149,
        4150,
        4151,
        4152,
        4153,
        4154,
        4155,
        4156,
        4157,
        4158,
        4159,
        4160,
        4161,
        4162,
        4163,
        4164,
        4165,
        4166,
        4167,
        4168,
        4169,
        4170,
        4171,
        4172,
        4173,
        4174,
        4175,
        4176,
        4177,
        4178,
        4179,
        4180,
        4181,
        5001,
        5002,
        5100,
        5101,
        5102,
        5103,
        5104,
        5105,
        5106,
        5107,
        5108,
        5109,
        5110,
        5111,
        5112,
        5113,
        5114,
        5115,
        5116,
        5117,
        5200,
        5201,
        5202,
        5203,
        5204,
        5205,
        5206,
        5207,
        5208,
        5209,
        5210,
        5211,
        5212,
        5213,
        5214,
        5215,
        5216,
        5217,
        6000,
        6001,
        6002,
        6003,
        6010,
        6020,
        6030,
        6040,
        6050,
        6060,
        6070,
        6080,
        7001,
        7002,
        7003,
        7004,
        7005,
        7006,
        7007,
        7008,
        7009,
        7010,
        7011,
        7012,
        7013,
        7014,
        7015,
        7016,
        7017,
        7018,
        7019,
        7020,
        7100,
        7101,
        8001,
        8002,
        8003,
        8004,
        8005,
        8006,
        8007,
        8008,
        8009,
        8010,
        8011,
        8012,
        8013,
        8014,
        9000,
        9001,
        9002,
        9003,
        9004,
        9005,
        9006,
        9500,
        9501,
        9502,
        9503,
        9504,
        9505,
        9506,
        11100,
        11101,
        11102,
        11103,
        11104,
        11105,
        11106,
        11107,
        11108,
        11109,
        11110,
        11111,
        11112,
        11113,
        11114,
        11115,
        11116,
        11117,
        11118,
        11119,
        11120,
        11121,
        11122,
        11123,
        11124,
        11125,
        11126,
        11127,
        11128,
        11129,
        11130,
        11131,
        11132,
        11133,
        11134,
        11135,
        11136,
        11137,
        11138,
        11139,
        11140,
        11141,
        11142,
        11143,
        11144,
        11145,
        11146,
        11147,
        11148,
        11149,
        11150,
        11151,
        11152,
        11153,
        11154,
        11155,
        11156,
        11157,
        11158,
        11159,
        11160,
        11161,
        11162,
        11163,
        11164,
        11165,
        11166,
        11167,
        11168,
        11169,
        11170,
        11171,
        11172,
        11173,
        11174,
        11175,
        12100,
        12101,
        12102,
        12103,
        12104,
        12105,
        12106,
        12107,
        12108,
        12109,
        12110,
        12111,
        12112,
        12113,
        12114,
        12115,
        12116,
        12117,
        12118,
        12119,
        12120,
        12121,
        12122,
        12123,
        12124,
        12125,
        12126,
        12127,
        12128,
        12129,
        12130,
        12131,
        12132,
        12133,
        12134,
        12135,
        12136,
        12137,
        12138,
        12139,
        12140,
        12141,
        12142,
        12143,
        12144,
        12145,
        12146,
        12147,
        12148,
        12149,
        12150,
        12151,
        12152,
        12153,
        12154,
        12155,
        12156,
        12157,
        12158,
        12159,
        12160,
        12161,
        12162,
        12163,
        12164,
        12165,
        12166,
        12167,
        12168,
        12169,
        12170,
        12171,
        12172,
        12173,
        12174,
        12175,
        13100,
        13101,
        13102,
        13103,
        13104,
        13105,
        13106,
        13107,
        13108,
        13109,
        13110,
        13111,
        13112,
        13113,
        13114,
        13115,
        13116,
        13117,
        13118,
        13119,
        13120,
        13121,
        13122,
        13123,
        13124,
        13125,
        13126,
        13127,
        13128,
        13129,
        13130,
        13131,
        13132,
        13133,
        13134,
        13135,
        13136,
        13137,
        13138,
        13139,
        13140,
        13141,
        13142,
        13143,
        13144,
        13145,
        13146,
        13147,
        13148,
        13149,
        13150,
        13151,
        13152,
        13153,
        13154,
        13155,
        13156,
        13157,
        13158,
        13159,
        13160,
        13161,
        13162,
        13163,
        13164,
        13165,
        13166,
        13167,
        13168,
        13169,
        13170,
        13171,
        13172,
        13173,
        13174,
        13175,
        14100,
        14101,
        14102,
        14103,
        14104,
        14105,
        14106,
        14107,
        14108,
        14109,
        14110,
        14111,
        14112,
        14113,
        14114,
        14115,
        14116,
        14117,
        14118,
        14119,
        14120,
        14121,
        14122,
        14123,
        14124,
        14125,
        14126,
        14127,
        14128,
        14129,
        14130,
        14131,
        14132,
        14133,
        14134,
        14135,
        14136,
        14137,
        14138,
        14139,
        14140,
        14141,
        14142,
        14143,
        14144,
        14145,
        14146,
        14147,
        14148,
        14149,
        14150,
        14151,
        14152,
        14153,
        14154,
        14155,
        14156,
        14157,
        14158,
        14159,
        14160,
        14161,
        14162,
        14163,
        14164,
        14165,
        14166,
        14167,
        14168,
        14169,
        14170,
        14171,
        14172,
        14173,
        14174,
        14175]

    names = ['Unknown',
        'Left-Cerebral-Exterior',
        'Left-Cerebral-White-Matter',
        'Left-Cerebral-Cortex',
        'Left-Lateral-Ventricle',
        'Left-Inf-Lat-Vent',
        'Left-Cerebellum-Exterior',
        'Left-Cerebellum-White-Matter',
        'Left-Cerebellum-Cortex',
        'Left-Thalamus',
        'Left-Thalamus-Proper',
        'Left-Caudate',
        'Left-Putamen',
        'Left-Pallidum',
        '3rd-Ventricle',
        '4th-Ventricle',
        'Brain-Stem',
        'Left-Hippocampus',
        'Left-Amygdala',
        'Left-Insula',
        'Left-Operculum',
        'Line-1',
        'Line-2',
        'Line-3',
        'CSF',
        'Left-Lesion',
        'Left-Accumbens-area',
        'Left-Substancia-Nigra',
        'Left-VentralDC',
        'Left-undetermined',
        'Left-vessel',
        'Left-choroid-plexus',
        'Left-F3orb',
        'Left-lOg',
        'Left-aOg',
        'Left-mOg',
        'Left-pOg',
        'Left-Stellate',
        'Left-Porg',
        'Left-Aorg',
        'Right-Cerebral-Exterior',
        'Right-Cerebral-White-Matter',
        'Right-Cerebral-Cortex',
        'Right-Lateral-Ventricle',
        'Right-Inf-Lat-Vent',
        'Right-Cerebellum-Exterior',
        'Right-Cerebellum-White-Matter',
        'Right-Cerebellum-Cortex',
        'Right-Thalamus',
        'Right-Thalamus-Proper',
        'Right-Caudate',
        'Right-Putamen',
        'Right-Pallidum',
        'Right-Hippocampus',
        'Right-Amygdala',
        'Right-Insula',
        'Right-Operculum',
        'Right-Lesion',
        'Right-Accumbens-area',
        'Right-Substancia-Nigra',
        'Right-VentralDC',
        'Right-undetermined',
        'Right-vessel',
        'Right-choroid-plexus',
        'Right-F3orb',
        'Right-lOg',
        'Right-aOg',
        'Right-mOg',
        'Right-pOg',
        'Right-Stellate',
        'Right-Porg',
        'Right-Aorg',
        '5th-Ventricle',
        'Left-Interior',
        'Right-Interior',
        'WM-hypointensities',
        'Left-WM-hypointensities',
        'Right-WM-hypointensities',
        'non-WM-hypointensities',
        'Left-non-WM-hypointensities',
        'Right-non-WM-hypointensities',
        'Left-F1',
        'Right-F1',
        'Optic-Chiasm',
        'Corpus_Callosum',
        'Left_future_WMSA',
        'Right_future_WMSA',
        'future_WMSA',
        'Left-Amygdala-Anterior',
        'Right-Amygdala-Anterior',
        'Dura',
        'Left-wm-intensity-abnormality',
        'Left-caudate-intensity-abnormality',
        'Left-putamen-intensity-abnormality',
        'Left-accumbens-intensity-abnormality',
        'Left-pallidum-intensity-abnormality',
        'Left-amygdala-intensity-abnormality',
        'Left-hippocampus-intensity-abnormality',
        'Left-thalamus-intensity-abnormality',
        'Left-VDC-intensity-abnormality',
        'Right-wm-intensity-abnormality',
        'Right-caudate-intensity-abnormality',
        'Right-putamen-intensity-abnormality',
        'Right-accumbens-intensity-abnormality',
        'Right-pallidum-intensity-abnormality',
        'Right-amygdala-intensity-abnormality',
        'Right-hippocampus-intensity-abnormality',
        'Right-thalamus-intensity-abnormality',
        'Right-VDC-intensity-abnormality',
        'Epidermis',
        'Conn-Tissue',
        'SC-Fat-Muscle',
        'Cranium',
        'CSF-SA',
        'Muscle',
        'Ear',
        'Adipose',
        'Spinal-Cord',
        'Soft-Tissue',
        'Nerve',
        'Bone',
        'Air',
        'Orbital-Fat',
        'Tongue',
        'Nasal-Structures',
        'Globe',
        'Teeth',
        'Left-Caudate-Putamen',
        'Right-Caudate-Putamen',
        'Left-Claustrum',
        'Right-Claustrum',
        'Cornea',
        'Diploe',
        'Vitreous-Humor',
        'Lens',
        'Aqueous-Humor',
        'Outer-Table',
        'Inner-Table',
        'Periosteum',
        'Endosteum',
        'R-C-S',
        'Iris',
        'SC-Adipose-Muscle',
        'SC-Tissue',
        'Orbital-Adipose',
        'Left-IntCapsule-Ant',
        'Right-IntCapsule-Ant',
        'Left-IntCapsule-Pos',
        'Right-IntCapsule-Pos',
        'Left-Cerebral-WM-unmyelinated',
        'Right-Cerebral-WM-unmyelinated',
        'Left-Cerebral-WM-myelinated',
        'Right-Cerebral-WM-myelinated',
        'Left-Subcortical-Gray-Matter',
        'Right-Subcortical-Gray-Matter',
        'Skull',
        'Posterior-fossa',
        'Scalp',
        'Hematoma',
        'Left-Basal-Ganglia',
        'Right-Basal-Ganglia',
        'brainstem',
        'DCG',
        'Vermis',
        'Midbrain',
        'Pons',
        'Medulla',
        'Left-Cortical-Dysplasia',
        'Right-Cortical-Dysplasia',
        'Left-hippocampal_fissure',
        'Left-CADG-head',
        'Left-subiculum',
        'Left-fimbria',
        'Right-hippocampal_fissure',
        'Right-CADG-head',
        'Right-subiculum',
        'Right-fimbria',
        'alveus',
        'perforant_pathway',
        'parasubiculum',
        'presubiculum',
        'subiculum',
        'CA1',
        'CA2',
        'CA3',
        'CA4',
        'GC-ML-DG',
        'HATA',
        'fimbria',
        'lateral_ventricle',
        'molecular_layer_HP',
        'hippocampal_fissure',
        'entorhinal_cortex',
        'molecular_layer_subiculum',
        'Amygdala',
        'Cerebral_White_Matter',
        'Cerebral_Cortex',
        'Inf_Lat_Vent',
        'Perirhinal',
        'Cerebral_White_Matter_Edge',
        'Background',
        'Ectorhinal',
        'HP_tail',
        'Fornix',
        'CC_Posterior',
        'CC_Mid_Posterior',
        'CC_Central',
        'CC_Mid_Anterior',
        'CC_Anterior',
        'Voxel-Unchanged',
        'Aorta',
        'Left-Common-IliacA',
        'Right-Common-IliacA',
        'Left-External-IliacA',
        'Right-External-IliacA',
        'Left-Internal-IliacA',
        'Right-Internal-IliacA',
        'Left-Lateral-SacralA',
        'Right-Lateral-SacralA',
        'Left-ObturatorA',
        'Right-ObturatorA',
        'Left-Internal-PudendalA',
        'Right-Internal-PudendalA',
        'Left-UmbilicalA',
        'Right-UmbilicalA',
        'Left-Inf-RectalA',
        'Right-Inf-RectalA',
        'Left-Common-IliacV',
        'Right-Common-IliacV',
        'Left-External-IliacV',
        'Right-External-IliacV',
        'Left-Internal-IliacV',
        'Right-Internal-IliacV',
        'Left-ObturatorV',
        'Right-ObturatorV',
        'Left-Internal-PudendalV',
        'Right-Internal-PudendalV',
        'Pos-Lymph',
        'Neg-Lymph',
        'V1',
        'V2',
        'BA44',
        'BA45',
        'BA4a',
        'BA4p',
        'BA6',
        'BA2',
        'BA1_old',
        'BAun2',
        'BA1',
        'BA2b',
        'BA3a',
        'BA3b',
        'MT',
        'AIPS_AIP_l',
        'AIPS_AIP_r',
        'AIPS_VIP_l',
        'AIPS_VIP_r',
        'IPL_PFcm_l',
        'IPL_PFcm_r',
        'IPL_PF_l',
        'IPL_PFm_l',
        'IPL_PFm_r',
        'IPL_PFop_l',
        'IPL_PFop_r',
        'IPL_PF_r',
        'IPL_PFt_l',
        'IPL_PFt_r',
        'IPL_PGa_l',
        'IPL_PGa_r',
        'IPL_PGp_l',
        'IPL_PGp_r',
        'Visual_V3d_l',
        'Visual_V3d_r',
        'Visual_V4_l',
        'Visual_V4_r',
        'Visual_V5_b',
        'Visual_VP_l',
        'Visual_VP_r',
        'wmsa',
        'other_wmsa',
        'right_CA2_3',
        'right_alveus',
        'right_CA1',
        'right_fimbria',
        'right_presubiculum',
        'right_hippocampal_fissure',
        'right_CA4_DG',
        'right_subiculum',
        'right_fornix',
        'left_CA2_3',
        'left_alveus',
        'left_CA1',
        'left_fimbria',
        'left_presubiculum',
        'left_hippocampal_fissure',
        'left_CA4_DG',
        'left_subiculum',
        'left_fornix',
        'Tumor',
        'Cbm_Left_I_IV',
        'Cbm_Right_I_IV',
        'Cbm_Left_V',
        'Cbm_Right_V',
        'Cbm_Left_VI',
        'Cbm_Vermis_VI',
        'Cbm_Right_VI',
        'Cbm_Left_CrusI',
        'Cbm_Vermis_CrusI',
        'Cbm_Right_CrusI',
        'Cbm_Left_CrusII',
        'Cbm_Vermis_CrusII',
        'Cbm_Right_CrusII',
        'Cbm_Left_VIIb',
        'Cbm_Vermis_VIIb',
        'Cbm_Right_VIIb',
        'Cbm_Left_VIIIa',
        'Cbm_Vermis_VIIIa',
        'Cbm_Right_VIIIa',
        'Cbm_Left_VIIIb',
        'Cbm_Vermis_VIIIb',
        'Cbm_Right_VIIIb',
        'Cbm_Left_IX',
        'Cbm_Vermis_IX',
        'Cbm_Right_IX',
        'Cbm_Left_X',
        'Cbm_Vermis_X',
        'Cbm_Right_X',
        'Cbm_Right_I_V_med',
        'Cbm_Right_I_V_mid',
        'Cbm_Right_VI_med',
        'Cbm_Right_VI_mid',
        'Cbm_Right_VI_lat',
        'Cbm_Right_CrusI_med',
        'Cbm_Right_CrusI_mid',
        'Cbm_Right_CrusI_lat',
        'Cbm_Right_CrusII_med',
        'Cbm_Right_CrusII_mid',
        'Cbm_Right_CrusII_lat',
        'Cbm_Right_7med',
        'Cbm_Right_7mid',
        'Cbm_Right_7lat',
        'Cbm_Right_8med',
        'Cbm_Right_8mid',
        'Cbm_Right_8lat',
        'Cbm_Right_PUNs',
        'Cbm_Right_TONs',
        'Cbm_Right_FLOs',
        'Cbm_Left_I_V_med',
        'Cbm_Left_I_V_mid',
        'Cbm_Left_VI_med',
        'Cbm_Left_VI_mid',
        'Cbm_Left_VI_lat',
        'Cbm_Left_CrusI_med',
        'Cbm_Left_CrusI_mid',
        'Cbm_Left_CrusI_lat',
        'Cbm_Left_CrusII_med',
        'Cbm_Left_CrusII_mid',
        'Cbm_Left_CrusII_lat',
        'Cbm_Left_7med',
        'Cbm_Left_7mid',
        'Cbm_Left_7lat',
        'Cbm_Left_8med',
        'Cbm_Left_8mid',
        'Cbm_Left_8lat',
        'Cbm_Left_PUNs',
        'Cbm_Left_TONs',
        'Cbm_Left_FLOs',
        'CSF-FSL-FAST',
        'GrayMatter-FSL-FAST',
        'WhiteMatter-FSL-FAST',
        'SUSPICIOUS',
        'ctx-lh-unknown',
        'ctx-lh-bankssts',
        'ctx-lh-caudalanteriorcingulate',
        'ctx-lh-caudalmiddlefrontal',
        'ctx-lh-corpuscallosum',
        'ctx-lh-cuneus',
        'ctx-lh-entorhinal',
        'ctx-lh-fusiform',
        'ctx-lh-inferiorparietal',
        'ctx-lh-inferiortemporal',
        'ctx-lh-isthmuscingulate',
        'ctx-lh-lateraloccipital',
        'ctx-lh-lateralorbitofrontal',
        'ctx-lh-lingual',
        'ctx-lh-medialorbitofrontal',
        'ctx-lh-middletemporal',
        'ctx-lh-parahippocampal',
        'ctx-lh-paracentral',
        'ctx-lh-parsopercularis',
        'ctx-lh-parsorbitalis',
        'ctx-lh-parstriangularis',
        'ctx-lh-pericalcarine',
        'ctx-lh-postcentral',
        'ctx-lh-posteriorcingulate',
        'ctx-lh-precentral',
        'ctx-lh-precuneus',
        'ctx-lh-rostralanteriorcingulate',
        'ctx-lh-rostralmiddlefrontal',
        'ctx-lh-superiorfrontal',
        'ctx-lh-superiorparietal',
        'ctx-lh-superiortemporal',
        'ctx-lh-supramarginal',
        'ctx-lh-frontalpole',
        'ctx-lh-temporalpole',
        'ctx-lh-transversetemporal',
        'ctx-lh-insula',
        'ctx-rh-unknown',
        'ctx-rh-bankssts',
        'ctx-rh-caudalanteriorcingulate',
        'ctx-rh-caudalmiddlefrontal',
        'ctx-rh-corpuscallosum',
        'ctx-rh-cuneus',
        'ctx-rh-entorhinal',
        'ctx-rh-fusiform',
        'ctx-rh-inferiorparietal',
        'ctx-rh-inferiortemporal',
        'ctx-rh-isthmuscingulate',
        'ctx-rh-lateraloccipital',
        'ctx-rh-lateralorbitofrontal',
        'ctx-rh-lingual',
        'ctx-rh-medialorbitofrontal',
        'ctx-rh-middletemporal',
        'ctx-rh-parahippocampal',
        'ctx-rh-paracentral',
        'ctx-rh-parsopercularis',
        'ctx-rh-parsorbitalis',
        'ctx-rh-parstriangularis',
        'ctx-rh-pericalcarine',
        'ctx-rh-postcentral',
        'ctx-rh-posteriorcingulate',
        'ctx-rh-precentral',
        'ctx-rh-precuneus',
        'ctx-rh-rostralanteriorcingulate',
        'ctx-rh-rostralmiddlefrontal',
        'ctx-rh-superiorfrontal',
        'ctx-rh-superiorparietal',
        'ctx-rh-superiortemporal',
        'ctx-rh-supramarginal',
        'ctx-rh-frontalpole',
        'ctx-rh-temporalpole',
        'ctx-rh-transversetemporal',
        'ctx-rh-insula',
        'wm-lh-unknown',
        'wm-lh-bankssts',
        'wm-lh-caudalanteriorcingulate',
        'wm-lh-caudalmiddlefrontal',
        'wm-lh-corpuscallosum',
        'wm-lh-cuneus',
        'wm-lh-entorhinal',
        'wm-lh-fusiform',
        'wm-lh-inferiorparietal',
        'wm-lh-inferiortemporal',
        'wm-lh-isthmuscingulate',
        'wm-lh-lateraloccipital',
        'wm-lh-lateralorbitofrontal',
        'wm-lh-lingual',
        'wm-lh-medialorbitofrontal',
        'wm-lh-middletemporal',
        'wm-lh-parahippocampal',
        'wm-lh-paracentral',
        'wm-lh-parsopercularis',
        'wm-lh-parsorbitalis',
        'wm-lh-parstriangularis',
        'wm-lh-pericalcarine',
        'wm-lh-postcentral',
        'wm-lh-posteriorcingulate',
        'wm-lh-precentral',
        'wm-lh-precuneus',
        'wm-lh-rostralanteriorcingulate',
        'wm-lh-rostralmiddlefrontal',
        'wm-lh-superiorfrontal',
        'wm-lh-superiorparietal',
        'wm-lh-superiortemporal',
        'wm-lh-supramarginal',
        'wm-lh-frontalpole',
        'wm-lh-temporalpole',
        'wm-lh-transversetemporal',
        'wm-lh-insula',
        'wm-rh-unknown',
        'wm-rh-bankssts',
        'wm-rh-caudalanteriorcingulate',
        'wm-rh-caudalmiddlefrontal',
        'wm-rh-corpuscallosum',
        'wm-rh-cuneus',
        'wm-rh-entorhinal',
        'wm-rh-fusiform',
        'wm-rh-inferiorparietal',
        'wm-rh-inferiortemporal',
        'wm-rh-isthmuscingulate',
        'wm-rh-lateraloccipital',
        'wm-rh-lateralorbitofrontal',
        'wm-rh-lingual',
        'wm-rh-medialorbitofrontal',
        'wm-rh-middletemporal',
        'wm-rh-parahippocampal',
        'wm-rh-paracentral',
        'wm-rh-parsopercularis',
        'wm-rh-parsorbitalis',
        'wm-rh-parstriangularis',
        'wm-rh-pericalcarine',
        'wm-rh-postcentral',
        'wm-rh-posteriorcingulate',
        'wm-rh-precentral',
        'wm-rh-precuneus',
        'wm-rh-rostralanteriorcingulate',
        'wm-rh-rostralmiddlefrontal',
        'wm-rh-superiorfrontal',
        'wm-rh-superiorparietal',
        'wm-rh-superiortemporal',
        'wm-rh-supramarginal',
        'wm-rh-frontalpole',
        'wm-rh-temporalpole',
        'wm-rh-transversetemporal',
        'wm-rh-insula',
        'ctx-lh-Unknown',
        'ctx-lh-Corpus_callosum',
        'ctx-lh-G_and_S_Insula_ONLY_AVERAGE',
        'ctx-lh-G_cingulate-Isthmus',
        'ctx-lh-G_cingulate-Main_part',
        'ctx-lh-G_cingulate-caudal_ACC',
        'ctx-lh-G_cingulate-rostral_ACC',
        'ctx-lh-G_cingulate-posterior',
        'ctx-lh-S_cingulate-caudal_ACC',
        'ctx-lh-S_cingulate-rostral_ACC',
        'ctx-lh-S_cingulate-posterior',
        'ctx-lh-S_pericallosal-caudal',
        'ctx-lh-S_pericallosal-rostral',
        'ctx-lh-S_pericallosal-posterior',
        'ctx-lh-G_cuneus',
        'ctx-lh-G_frontal_inf-Opercular_part',
        'ctx-lh-G_frontal_inf-Orbital_part',
        'ctx-lh-G_frontal_inf-Triangular_part',
        'ctx-lh-G_frontal_middle',
        'ctx-lh-G_frontal_superior',
        'ctx-lh-G_frontomarginal',
        'ctx-lh-G_insular_long',
        'ctx-lh-G_insular_short',
        'ctx-lh-G_and_S_occipital_inferior',
        'ctx-lh-G_occipital_middle',
        'ctx-lh-G_occipital_superior',
        'ctx-lh-G_occipit-temp_lat-Or_fusiform',
        'ctx-lh-G_occipit-temp_med-Lingual_part',
        'ctx-lh-G_occipit-temp_med-Parahippocampal_part',
        'ctx-lh-G_orbital',
        'ctx-lh-G_paracentral',
        'ctx-lh-G_parietal_inferior-Angular_part',
        'ctx-lh-G_parietal_inferior-Supramarginal_part',
        'ctx-lh-G_parietal_superior',
        'ctx-lh-G_postcentral',
        'ctx-lh-G_precentral',
        'ctx-lh-G_precuneus',
        'ctx-lh-G_rectus',
        'ctx-lh-G_subcallosal',
        'ctx-lh-G_subcentral',
        'ctx-lh-G_temporal_inferior',
        'ctx-lh-G_temporal_middle',
        'ctx-lh-G_temp_sup-G_temp_transv_and_interm_S',
        'ctx-lh-G_temp_sup-Lateral_aspect',
        'ctx-lh-G_temp_sup-Planum_polare',
        'ctx-lh-G_temp_sup-Planum_tempolare',
        'ctx-lh-G_and_S_transverse_frontopolar',
        'ctx-lh-Lat_Fissure-ant_sgt-ramus_horizontal',
        'ctx-lh-Lat_Fissure-ant_sgt-ramus_vertical',
        'ctx-lh-Lat_Fissure-post_sgt',
        'ctx-lh-Medial_wall',
        'ctx-lh-Pole_occipital',
        'ctx-lh-Pole_temporal',
        'ctx-lh-S_calcarine',
        'ctx-lh-S_central',
        'ctx-lh-S_central_insula',
        'ctx-lh-S_cingulate-Main_part_and_Intracingulate',
        'ctx-lh-S_cingulate-Marginalis_part',
        'ctx-lh-S_circular_insula_anterior',
        'ctx-lh-S_circular_insula_inferior',
        'ctx-lh-S_circular_insula_superior',
        'ctx-lh-S_collateral_transverse_ant',
        'ctx-lh-S_collateral_transverse_post',
        'ctx-lh-S_frontal_inferior',
        'ctx-lh-S_frontal_middle',
        'ctx-lh-S_frontal_superior',
        'ctx-lh-S_frontomarginal',
        'ctx-lh-S_intermedius_primus-Jensen',
        'ctx-lh-S_intraparietal-and_Parietal_transverse',
        'ctx-lh-S_occipital_anterior',
        'ctx-lh-S_occipital_middle_and_Lunatus',
        'ctx-lh-S_occipital_superior_and_transversalis',
        'ctx-lh-S_occipito-temporal_lateral',
        'ctx-lh-S_occipito-temporal_medial_and_S_Lingual',
        'ctx-lh-S_orbital-H_shapped',
        'ctx-lh-S_orbital_lateral',
        'ctx-lh-S_orbital_medial-Or_olfactory',
        'ctx-lh-S_paracentral',
        'ctx-lh-S_parieto_occipital',
        'ctx-lh-S_pericallosal',
        'ctx-lh-S_postcentral',
        'ctx-lh-S_precentral-Inferior-part',
        'ctx-lh-S_precentral-Superior-part',
        'ctx-lh-S_subcentral_ant',
        'ctx-lh-S_subcentral_post',
        'ctx-lh-S_suborbital',
        'ctx-lh-S_subparietal',
        'ctx-lh-S_supracingulate',
        'ctx-lh-S_temporal_inferior',
        'ctx-lh-S_temporal_superior',
        'ctx-lh-S_temporal_transverse',
        'ctx-rh-Unknown',
        'ctx-rh-Corpus_callosum',
        'ctx-rh-G_and_S_Insula_ONLY_AVERAGE',
        'ctx-rh-G_cingulate-Isthmus',
        'ctx-rh-G_cingulate-Main_part',
        'ctx-rh-G_cuneus',
        'ctx-rh-G_frontal_inf-Opercular_part',
        'ctx-rh-G_frontal_inf-Orbital_part',
        'ctx-rh-G_frontal_inf-Triangular_part',
        'ctx-rh-G_frontal_middle',
        'ctx-rh-G_frontal_superior',
        'ctx-rh-G_frontomarginal',
        'ctx-rh-G_insular_long',
        'ctx-rh-G_insular_short',
        'ctx-rh-G_and_S_occipital_inferior',
        'ctx-rh-G_occipital_middle',
        'ctx-rh-G_occipital_superior',
        'ctx-rh-G_occipit-temp_lat-Or_fusiform',
        'ctx-rh-G_occipit-temp_med-Lingual_part',
        'ctx-rh-G_occipit-temp_med-Parahippocampal_part',
        'ctx-rh-G_orbital',
        'ctx-rh-G_paracentral',
        'ctx-rh-G_parietal_inferior-Angular_part',
        'ctx-rh-G_parietal_inferior-Supramarginal_part',
        'ctx-rh-G_parietal_superior',
        'ctx-rh-G_postcentral',
        'ctx-rh-G_precentral',
        'ctx-rh-G_precuneus',
        'ctx-rh-G_rectus',
        'ctx-rh-G_subcallosal',
        'ctx-rh-G_subcentral',
        'ctx-rh-G_temporal_inferior',
        'ctx-rh-G_temporal_middle',
        'ctx-rh-G_temp_sup-G_temp_transv_and_interm_S',
        'ctx-rh-G_temp_sup-Lateral_aspect',
        'ctx-rh-G_temp_sup-Planum_polare',
        'ctx-rh-G_temp_sup-Planum_tempolare',
        'ctx-rh-G_and_S_transverse_frontopolar',
        'ctx-rh-Lat_Fissure-ant_sgt-ramus_horizontal',
        'ctx-rh-Lat_Fissure-ant_sgt-ramus_vertical',
        'ctx-rh-Lat_Fissure-post_sgt',
        'ctx-rh-Medial_wall',
        'ctx-rh-Pole_occipital',
        'ctx-rh-Pole_temporal',
        'ctx-rh-S_calcarine',
        'ctx-rh-S_central',
        'ctx-rh-S_central_insula',
        'ctx-rh-S_cingulate-Main_part_and_Intracingulate',
        'ctx-rh-S_cingulate-Marginalis_part',
        'ctx-rh-S_circular_insula_anterior',
        'ctx-rh-S_circular_insula_inferior',
        'ctx-rh-S_circular_insula_superior',
        'ctx-rh-S_collateral_transverse_ant',
        'ctx-rh-S_collateral_transverse_post',
        'ctx-rh-S_frontal_inferior',
        'ctx-rh-S_frontal_middle',
        'ctx-rh-S_frontal_superior',
        'ctx-rh-S_frontomarginal',
        'ctx-rh-S_intermedius_primus-Jensen',
        'ctx-rh-S_intraparietal-and_Parietal_transverse',
        'ctx-rh-S_occipital_anterior',
        'ctx-rh-S_occipital_middle_and_Lunatus',
        'ctx-rh-S_occipital_superior_and_transversalis',
        'ctx-rh-S_occipito-temporal_lateral',
        'ctx-rh-S_occipito-temporal_medial_and_S_Lingual',
        'ctx-rh-S_orbital-H_shapped',
        'ctx-rh-S_orbital_lateral',
        'ctx-rh-S_orbital_medial-Or_olfactory',
        'ctx-rh-S_paracentral',
        'ctx-rh-S_parieto_occipital',
        'ctx-rh-S_pericallosal',
        'ctx-rh-S_postcentral',
        'ctx-rh-S_precentral-Inferior-part',
        'ctx-rh-S_precentral-Superior-part',
        'ctx-rh-S_subcentral_ant',
        'ctx-rh-S_subcentral_post',
        'ctx-rh-S_suborbital',
        'ctx-rh-S_subparietal',
        'ctx-rh-S_supracingulate',
        'ctx-rh-S_temporal_inferior',
        'ctx-rh-S_temporal_superior',
        'ctx-rh-S_temporal_transverse',
        'ctx-rh-G_cingulate-caudal_ACC',
        'ctx-rh-G_cingulate-rostral_ACC',
        'ctx-rh-G_cingulate-posterior',
        'ctx-rh-S_cingulate-caudal_ACC',
        'ctx-rh-S_cingulate-rostral_ACC',
        'ctx-rh-S_cingulate-posterior',
        'ctx-rh-S_pericallosal-caudal',
        'ctx-rh-S_pericallosal-rostral',
        'ctx-rh-S_pericallosal-posterior',
        'wm-lh-Unknown',
        'wm-lh-Corpus_callosum',
        'wm-lh-G_and_S_Insula_ONLY_AVERAGE',
        'wm-lh-G_cingulate-Isthmus',
        'wm-lh-G_cingulate-Main_part',
        'wm-lh-G_cuneus',
        'wm-lh-G_frontal_inf-Opercular_part',
        'wm-lh-G_frontal_inf-Orbital_part',
        'wm-lh-G_frontal_inf-Triangular_part',
        'wm-lh-G_frontal_middle',
        'wm-lh-G_frontal_superior',
        'wm-lh-G_frontomarginal',
        'wm-lh-G_insular_long',
        'wm-lh-G_insular_short',
        'wm-lh-G_and_S_occipital_inferior',
        'wm-lh-G_occipital_middle',
        'wm-lh-G_occipital_superior',
        'wm-lh-G_occipit-temp_lat-Or_fusiform',
        'wm-lh-G_occipit-temp_med-Lingual_part',
        'wm-lh-G_occipit-temp_med-Parahippocampal_part',
        'wm-lh-G_orbital',
        'wm-lh-G_paracentral',
        'wm-lh-G_parietal_inferior-Angular_part',
        'wm-lh-G_parietal_inferior-Supramarginal_part',
        'wm-lh-G_parietal_superior',
        'wm-lh-G_postcentral',
        'wm-lh-G_precentral',
        'wm-lh-G_precuneus',
        'wm-lh-G_rectus',
        'wm-lh-G_subcallosal',
        'wm-lh-G_subcentral',
        'wm-lh-G_temporal_inferior',
        'wm-lh-G_temporal_middle',
        'wm-lh-G_temp_sup-G_temp_transv_and_interm_S',
        'wm-lh-G_temp_sup-Lateral_aspect',
        'wm-lh-G_temp_sup-Planum_polare',
        'wm-lh-G_temp_sup-Planum_tempolare',
        'wm-lh-G_and_S_transverse_frontopolar',
        'wm-lh-Lat_Fissure-ant_sgt-ramus_horizontal',
        'wm-lh-Lat_Fissure-ant_sgt-ramus_vertical',
        'wm-lh-Lat_Fissure-post_sgt',
        'wm-lh-Medial_wall',
        'wm-lh-Pole_occipital',
        'wm-lh-Pole_temporal',
        'wm-lh-S_calcarine',
        'wm-lh-S_central',
        'wm-lh-S_central_insula',
        'wm-lh-S_cingulate-Main_part_and_Intracingulate',
        'wm-lh-S_cingulate-Marginalis_part',
        'wm-lh-S_circular_insula_anterior',
        'wm-lh-S_circular_insula_inferior',
        'wm-lh-S_circular_insula_superior',
        'wm-lh-S_collateral_transverse_ant',
        'wm-lh-S_collateral_transverse_post',
        'wm-lh-S_frontal_inferior',
        'wm-lh-S_frontal_middle',
        'wm-lh-S_frontal_superior',
        'wm-lh-S_frontomarginal',
        'wm-lh-S_intermedius_primus-Jensen',
        'wm-lh-S_intraparietal-and_Parietal_transverse',
        'wm-lh-S_occipital_anterior',
        'wm-lh-S_occipital_middle_and_Lunatus',
        'wm-lh-S_occipital_superior_and_transversalis',
        'wm-lh-S_occipito-temporal_lateral',
        'wm-lh-S_occipito-temporal_medial_and_S_Lingual',
        'wm-lh-S_orbital-H_shapped',
        'wm-lh-S_orbital_lateral',
        'wm-lh-S_orbital_medial-Or_olfactory',
        'wm-lh-S_paracentral',
        'wm-lh-S_parieto_occipital',
        'wm-lh-S_pericallosal',
        'wm-lh-S_postcentral',
        'wm-lh-S_precentral-Inferior-part',
        'wm-lh-S_precentral-Superior-part',
        'wm-lh-S_subcentral_ant',
        'wm-lh-S_subcentral_post',
        'wm-lh-S_suborbital',
        'wm-lh-S_subparietal',
        'wm-lh-S_supracingulate',
        'wm-lh-S_temporal_inferior',
        'wm-lh-S_temporal_superior',
        'wm-lh-S_temporal_transverse',
        'wm-rh-Unknown',
        'wm-rh-Corpus_callosum',
        'wm-rh-G_and_S_Insula_ONLY_AVERAGE',
        'wm-rh-G_cingulate-Isthmus',
        'wm-rh-G_cingulate-Main_part',
        'wm-rh-G_cuneus',
        'wm-rh-G_frontal_inf-Opercular_part',
        'wm-rh-G_frontal_inf-Orbital_part',
        'wm-rh-G_frontal_inf-Triangular_part',
        'wm-rh-G_frontal_middle',
        'wm-rh-G_frontal_superior',
        'wm-rh-G_frontomarginal',
        'wm-rh-G_insular_long',
        'wm-rh-G_insular_short',
        'wm-rh-G_and_S_occipital_inferior',
        'wm-rh-G_occipital_middle',
        'wm-rh-G_occipital_superior',
        'wm-rh-G_occipit-temp_lat-Or_fusiform',
        'wm-rh-G_occipit-temp_med-Lingual_part',
        'wm-rh-G_occipit-temp_med-Parahippocampal_part',
        'wm-rh-G_orbital',
        'wm-rh-G_paracentral',
        'wm-rh-G_parietal_inferior-Angular_part',
        'wm-rh-G_parietal_inferior-Supramarginal_part',
        'wm-rh-G_parietal_superior',
        'wm-rh-G_postcentral',
        'wm-rh-G_precentral',
        'wm-rh-G_precuneus',
        'wm-rh-G_rectus',
        'wm-rh-G_subcallosal',
        'wm-rh-G_subcentral',
        'wm-rh-G_temporal_inferior',
        'wm-rh-G_temporal_middle',
        'wm-rh-G_temp_sup-G_temp_transv_and_interm_S',
        'wm-rh-G_temp_sup-Lateral_aspect',
        'wm-rh-G_temp_sup-Planum_polare',
        'wm-rh-G_temp_sup-Planum_tempolare',
        'wm-rh-G_and_S_transverse_frontopolar',
        'wm-rh-Lat_Fissure-ant_sgt-ramus_horizontal',
        'wm-rh-Lat_Fissure-ant_sgt-ramus_vertical',
        'wm-rh-Lat_Fissure-post_sgt',
        'wm-rh-Medial_wall',
        'wm-rh-Pole_occipital',
        'wm-rh-Pole_temporal',
        'wm-rh-S_calcarine',
        'wm-rh-S_central',
        'wm-rh-S_central_insula',
        'wm-rh-S_cingulate-Main_part_and_Intracingulate',
        'wm-rh-S_cingulate-Marginalis_part',
        'wm-rh-S_circular_insula_anterior',
        'wm-rh-S_circular_insula_inferior',
        'wm-rh-S_circular_insula_superior',
        'wm-rh-S_collateral_transverse_ant',
        'wm-rh-S_collateral_transverse_post',
        'wm-rh-S_frontal_inferior',
        'wm-rh-S_frontal_middle',
        'wm-rh-S_frontal_superior',
        'wm-rh-S_frontomarginal',
        'wm-rh-S_intermedius_primus-Jensen',
        'wm-rh-S_intraparietal-and_Parietal_transverse',
        'wm-rh-S_occipital_anterior',
        'wm-rh-S_occipital_middle_and_Lunatus',
        'wm-rh-S_occipital_superior_and_transversalis',
        'wm-rh-S_occipito-temporal_lateral',
        'wm-rh-S_occipito-temporal_medial_and_S_Lingual',
        'wm-rh-S_orbital-H_shapped',
        'wm-rh-S_orbital_lateral',
        'wm-rh-S_orbital_medial-Or_olfactory',
        'wm-rh-S_paracentral',
        'wm-rh-S_parieto_occipital',
        'wm-rh-S_pericallosal',
        'wm-rh-S_postcentral',
        'wm-rh-S_precentral-Inferior-part',
        'wm-rh-S_precentral-Superior-part',
        'wm-rh-S_subcentral_ant',
        'wm-rh-S_subcentral_post',
        'wm-rh-S_suborbital',
        'wm-rh-S_subparietal',
        'wm-rh-S_supracingulate',
        'wm-rh-S_temporal_inferior',
        'wm-rh-S_temporal_superior',
        'wm-rh-S_temporal_transverse',
        'Left-UnsegmentedWhiteMatter',
        'Right-UnsegmentedWhiteMatter',
        'fmajor',
        'fminor',
        'lh.atr',
        'lh.cab',
        'lh.ccg',
        'lh.cst',
        'lh.ilf',
        'lh.slfp',
        'lh.slft',
        'lh.unc',
        'rh.atr',
        'rh.cab',
        'rh.ccg',
        'rh.cst',
        'rh.ilf',
        'rh.slfp',
        'rh.slft',
        'rh.unc',
        'CC-ForcepsMajor',
        'CC-ForcepsMinor',
        'LAntThalRadiation',
        'LCingulumAngBundle',
        'LCingulumCingGyrus',
        'LCorticospinalTract',
        'LInfLongFas',
        'LSupLongFasParietal',
        'LSupLongFasTemporal',
        'LUncinateFas',
        'RAntThalRadiation',
        'RCingulumAngBundle',
        'RCingulumCingGyrus',
        'RCorticospinalTract',
        'RInfLongFas',
        'RSupLongFasParietal',
        'RSupLongFasTemporal',
        'RUncinateFas',
        'CST-orig',
        'CST-hammer',
        'CST-CVS',
        'CST-flirt',
        'Left-SLF1',
        'Right-SLF1',
        'Left-SLF3',
        'Right-SLF3',
        'Left-CST',
        'Right-CST',
        'Left-SLF2',
        'Right-SLF2',
        'Lateral-nucleus',
        'Basolateral-nucleus',
        'Basal-nucleus',
        'Centromedial-nucleus',
        'Central-nucleus',
        'Medial-nucleus',
        'Cortical-nucleus',
        'Accessory-Basal-nucleus',
        'Corticoamygdaloid-transitio',
        'Anterior-amygdaloid-area-AAA',
        'Fusion-amygdala-HP-FAH',
        'Hippocampal-amygdala-transition-HATA',
        'Endopiriform-nucleus',
        'Lateral-nucleus-olfactory-tract',
        'Paralaminar-nucleus',
        'Intercalated-nucleus',
        'Prepiriform-cortex',
        'Periamygdaloid-cortex',
        'Envelope-Amygdala',
        'Extranuclear-Amydala',
        'Brainstem-inferior-colliculus',
        'Brainstem-cochlear-nucleus',
        'Thalamus-Anterior',
        'Thalamus-Ventral-anterior',
        'Thalamus-Lateral-dorsal',
        'Thalamus-Lateral-posterior',
        'Thalamus-Ventral-lateral',
        'Thalamus-Ventral-posterior-medial',
        'Thalamus-Ventral-posterior-lateral',
        'Thalamus-intralaminar',
        'Thalamus-centromedian',
        'Thalamus-mediodorsal',
        'Thalamus-medial',
        'Thalamus-pulvinar',
        'Thalamus-lateral-geniculate',
        'Thalamus-medial-geniculate',
        'ctx-lh-prefrontal',
        'ctx-lh-primary-motor',
        'ctx-lh-premotor',
        'ctx-lh-temporal',
        'ctx-lh-posterior-parietal',
        'ctx-lh-prim-sec-somatosensory',
        'ctx-lh-occipital',
        'ctx-rh-prefrontal',
        'ctx-rh-primary-motor',
        'ctx-rh-premotor',
        'ctx-rh-temporal',
        'ctx-rh-posterior-parietal',
        'ctx-rh-prim-sec-somatosensory',
        'ctx-rh-occipital',
        'ctx_lh_Unknown',
        'ctx_lh_G_and_S_frontomargin',
        'ctx_lh_G_and_S_occipital_inf',
        'ctx_lh_G_and_S_paracentral',
        'ctx_lh_G_and_S_subcentral',
        'ctx_lh_G_and_S_transv_frontopol',
        'ctx_lh_G_and_S_cingul-Ant',
        'ctx_lh_G_and_S_cingul-Mid-Ant',
        'ctx_lh_G_and_S_cingul-Mid-Post',
        'ctx_lh_G_cingul-Post-dorsal',
        'ctx_lh_G_cingul-Post-ventral',
        'ctx_lh_G_cuneus',
        'ctx_lh_G_front_inf-Opercular',
        'ctx_lh_G_front_inf-Orbital',
        'ctx_lh_G_front_inf-Triangul',
        'ctx_lh_G_front_middle',
        'ctx_lh_G_front_sup',
        'ctx_lh_G_Ins_lg_and_S_cent_ins',
        'ctx_lh_G_insular_short',
        'ctx_lh_G_occipital_middle',
        'ctx_lh_G_occipital_sup',
        'ctx_lh_G_oc-temp_lat-fusifor',
        'ctx_lh_G_oc-temp_med-Lingual',
        'ctx_lh_G_oc-temp_med-Parahip',
        'ctx_lh_G_orbital',
        'ctx_lh_G_pariet_inf-Angular',
        'ctx_lh_G_pariet_inf-Supramar',
        'ctx_lh_G_parietal_sup',
        'ctx_lh_G_postcentral',
        'ctx_lh_G_precentral',
        'ctx_lh_G_precuneus',
        'ctx_lh_G_rectus',
        'ctx_lh_G_subcallosal',
        'ctx_lh_G_temp_sup-G_T_transv',
        'ctx_lh_G_temp_sup-Lateral',
        'ctx_lh_G_temp_sup-Plan_polar',
        'ctx_lh_G_temp_sup-Plan_tempo',
        'ctx_lh_G_temporal_inf',
        'ctx_lh_G_temporal_middle',
        'ctx_lh_Lat_Fis-ant-Horizont',
        'ctx_lh_Lat_Fis-ant-Vertical',
        'ctx_lh_Lat_Fis-post',
        'ctx_lh_Medial_wall',
        'ctx_lh_Pole_occipital',
        'ctx_lh_Pole_temporal',
        'ctx_lh_S_calcarine',
        'ctx_lh_S_central',
        'ctx_lh_S_cingul-Marginalis',
        'ctx_lh_S_circular_insula_ant',
        'ctx_lh_S_circular_insula_inf',
        'ctx_lh_S_circular_insula_sup',
        'ctx_lh_S_collat_transv_ant',
        'ctx_lh_S_collat_transv_post',
        'ctx_lh_S_front_inf',
        'ctx_lh_S_front_middle',
        'ctx_lh_S_front_sup',
        'ctx_lh_S_interm_prim-Jensen',
        'ctx_lh_S_intrapariet_and_P_trans',
        'ctx_lh_S_oc_middle_and_Lunatus',
        'ctx_lh_S_oc_sup_and_transversal',
        'ctx_lh_S_occipital_ant',
        'ctx_lh_S_oc-temp_lat',
        'ctx_lh_S_oc-temp_med_and_Lingual',
        'ctx_lh_S_orbital_lateral',
        'ctx_lh_S_orbital_med-olfact',
        'ctx_lh_S_orbital-H_Shaped',
        'ctx_lh_S_parieto_occipital',
        'ctx_lh_S_pericallosal',
        'ctx_lh_S_postcentral',
        'ctx_lh_S_precentral-inf-part',
        'ctx_lh_S_precentral-sup-part',
        'ctx_lh_S_suborbital',
        'ctx_lh_S_subparietal',
        'ctx_lh_S_temporal_inf',
        'ctx_lh_S_temporal_sup',
        'ctx_lh_S_temporal_transverse',
        'ctx_rh_Unknown',
        'ctx_rh_G_and_S_frontomargin',
        'ctx_rh_G_and_S_occipital_inf',
        'ctx_rh_G_and_S_paracentral',
        'ctx_rh_G_and_S_subcentral',
        'ctx_rh_G_and_S_transv_frontopol',
        'ctx_rh_G_and_S_cingul-Ant',
        'ctx_rh_G_and_S_cingul-Mid-Ant',
        'ctx_rh_G_and_S_cingul-Mid-Post',
        'ctx_rh_G_cingul-Post-dorsal',
        'ctx_rh_G_cingul-Post-ventral',
        'ctx_rh_G_cuneus',
        'ctx_rh_G_front_inf-Opercular',
        'ctx_rh_G_front_inf-Orbital',
        'ctx_rh_G_front_inf-Triangul',
        'ctx_rh_G_front_middle',
        'ctx_rh_G_front_sup',
        'ctx_rh_G_Ins_lg_and_S_cent_ins',
        'ctx_rh_G_insular_short',
        'ctx_rh_G_occipital_middle',
        'ctx_rh_G_occipital_sup',
        'ctx_rh_G_oc-temp_lat-fusifor',
        'ctx_rh_G_oc-temp_med-Lingual',
        'ctx_rh_G_oc-temp_med-Parahip',
        'ctx_rh_G_orbital',
        'ctx_rh_G_pariet_inf-Angular',
        'ctx_rh_G_pariet_inf-Supramar',
        'ctx_rh_G_parietal_sup',
        'ctx_rh_G_postcentral',
        'ctx_rh_G_precentral',
        'ctx_rh_G_precuneus',
        'ctx_rh_G_rectus',
        'ctx_rh_G_subcallosal',
        'ctx_rh_G_temp_sup-G_T_transv',
        'ctx_rh_G_temp_sup-Lateral',
        'ctx_rh_G_temp_sup-Plan_polar',
        'ctx_rh_G_temp_sup-Plan_tempo',
        'ctx_rh_G_temporal_inf',
        'ctx_rh_G_temporal_middle',
        'ctx_rh_Lat_Fis-ant-Horizont',
        'ctx_rh_Lat_Fis-ant-Vertical',
        'ctx_rh_Lat_Fis-post',
        'ctx_rh_Medial_wall',
        'ctx_rh_Pole_occipital',
        'ctx_rh_Pole_temporal',
        'ctx_rh_S_calcarine',
        'ctx_rh_S_central',
        'ctx_rh_S_cingul-Marginalis',
        'ctx_rh_S_circular_insula_ant',
        'ctx_rh_S_circular_insula_inf',
        'ctx_rh_S_circular_insula_sup',
        'ctx_rh_S_collat_transv_ant',
        'ctx_rh_S_collat_transv_post',
        'ctx_rh_S_front_inf',
        'ctx_rh_S_front_middle',
        'ctx_rh_S_front_sup',
        'ctx_rh_S_interm_prim-Jensen',
        'ctx_rh_S_intrapariet_and_P_trans',
        'ctx_rh_S_oc_middle_and_Lunatus',
        'ctx_rh_S_oc_sup_and_transversal',
        'ctx_rh_S_occipital_ant',
        'ctx_rh_S_oc-temp_lat',
        'ctx_rh_S_oc-temp_med_and_Lingual',
        'ctx_rh_S_orbital_lateral',
        'ctx_rh_S_orbital_med-olfact',
        'ctx_rh_S_orbital-H_Shaped',
        'ctx_rh_S_parieto_occipital',
        'ctx_rh_S_pericallosal',
        'ctx_rh_S_postcentral',
        'ctx_rh_S_precentral-inf-part',
        'ctx_rh_S_precentral-sup-part',
        'ctx_rh_S_suborbital',
        'ctx_rh_S_subparietal',
        'ctx_rh_S_temporal_inf',
        'ctx_rh_S_temporal_sup',
        'ctx_rh_S_temporal_transverse',
        'wm_lh_Unknown',
        'wm_lh_G_and_S_frontomargin',
        'wm_lh_G_and_S_occipital_inf',
        'wm_lh_G_and_S_paracentral',
        'wm_lh_G_and_S_subcentral',
        'wm_lh_G_and_S_transv_frontopol',
        'wm_lh_G_and_S_cingul-Ant',
        'wm_lh_G_and_S_cingul-Mid-Ant',
        'wm_lh_G_and_S_cingul-Mid-Post',
        'wm_lh_G_cingul-Post-dorsal',
        'wm_lh_G_cingul-Post-ventral',
        'wm_lh_G_cuneus',
        'wm_lh_G_front_inf-Opercular',
        'wm_lh_G_front_inf-Orbital',
        'wm_lh_G_front_inf-Triangul',
        'wm_lh_G_front_middle',
        'wm_lh_G_front_sup',
        'wm_lh_G_Ins_lg_and_S_cent_ins',
        'wm_lh_G_insular_short',
        'wm_lh_G_occipital_middle',
        'wm_lh_G_occipital_sup',
        'wm_lh_G_oc-temp_lat-fusifor',
        'wm_lh_G_oc-temp_med-Lingual',
        'wm_lh_G_oc-temp_med-Parahip',
        'wm_lh_G_orbital',
        'wm_lh_G_pariet_inf-Angular',
        'wm_lh_G_pariet_inf-Supramar',
        'wm_lh_G_parietal_sup',
        'wm_lh_G_postcentral',
        'wm_lh_G_precentral',
        'wm_lh_G_precuneus',
        'wm_lh_G_rectus',
        'wm_lh_G_subcallosal',
        'wm_lh_G_temp_sup-G_T_transv',
        'wm_lh_G_temp_sup-Lateral',
        'wm_lh_G_temp_sup-Plan_polar',
        'wm_lh_G_temp_sup-Plan_tempo',
        'wm_lh_G_temporal_inf',
        'wm_lh_G_temporal_middle',
        'wm_lh_Lat_Fis-ant-Horizont',
        'wm_lh_Lat_Fis-ant-Vertical',
        'wm_lh_Lat_Fis-post',
        'wm_lh_Medial_wall',
        'wm_lh_Pole_occipital',
        'wm_lh_Pole_temporal',
        'wm_lh_S_calcarine',
        'wm_lh_S_central',
        'wm_lh_S_cingul-Marginalis',
        'wm_lh_S_circular_insula_ant',
        'wm_lh_S_circular_insula_inf',
        'wm_lh_S_circular_insula_sup',
        'wm_lh_S_collat_transv_ant',
        'wm_lh_S_collat_transv_post',
        'wm_lh_S_front_inf',
        'wm_lh_S_front_middle',
        'wm_lh_S_front_sup',
        'wm_lh_S_interm_prim-Jensen',
        'wm_lh_S_intrapariet_and_P_trans',
        'wm_lh_S_oc_middle_and_Lunatus',
        'wm_lh_S_oc_sup_and_transversal',
        'wm_lh_S_occipital_ant',
        'wm_lh_S_oc-temp_lat',
        'wm_lh_S_oc-temp_med_and_Lingual',
        'wm_lh_S_orbital_lateral',
        'wm_lh_S_orbital_med-olfact',
        'wm_lh_S_orbital-H_Shaped',
        'wm_lh_S_parieto_occipital',
        'wm_lh_S_pericallosal',
        'wm_lh_S_postcentral',
        'wm_lh_S_precentral-inf-part',
        'wm_lh_S_precentral-sup-part',
        'wm_lh_S_suborbital',
        'wm_lh_S_subparietal',
        'wm_lh_S_temporal_inf',
        'wm_lh_S_temporal_sup',
        'wm_lh_S_temporal_transverse',
        'wm_rh_Unknown',
        'wm_rh_G_and_S_frontomargin',
        'wm_rh_G_and_S_occipital_inf',
        'wm_rh_G_and_S_paracentral',
        'wm_rh_G_and_S_subcentral',
        'wm_rh_G_and_S_transv_frontopol',
        'wm_rh_G_and_S_cingul-Ant',
        'wm_rh_G_and_S_cingul-Mid-Ant',
        'wm_rh_G_and_S_cingul-Mid-Post',
        'wm_rh_G_cingul-Post-dorsal',
        'wm_rh_G_cingul-Post-ventral',
        'wm_rh_G_cuneus',
        'wm_rh_G_front_inf-Opercular',
        'wm_rh_G_front_inf-Orbital',
        'wm_rh_G_front_inf-Triangul',
        'wm_rh_G_front_middle',
        'wm_rh_G_front_sup',
        'wm_rh_G_Ins_lg_and_S_cent_ins',
        'wm_rh_G_insular_short',
        'wm_rh_G_occipital_middle',
        'wm_rh_G_occipital_sup',
        'wm_rh_G_oc-temp_lat-fusifor',
        'wm_rh_G_oc-temp_med-Lingual',
        'wm_rh_G_oc-temp_med-Parahip',
        'wm_rh_G_orbital',
        'wm_rh_G_pariet_inf-Angular',
        'wm_rh_G_pariet_inf-Supramar',
        'wm_rh_G_parietal_sup',
        'wm_rh_G_postcentral',
        'wm_rh_G_precentral',
        'wm_rh_G_precuneus',
        'wm_rh_G_rectus',
        'wm_rh_G_subcallosal',
        'wm_rh_G_temp_sup-G_T_transv',
        'wm_rh_G_temp_sup-Lateral',
        'wm_rh_G_temp_sup-Plan_polar',
        'wm_rh_G_temp_sup-Plan_tempo',
        'wm_rh_G_temporal_inf',
        'wm_rh_G_temporal_middle',
        'wm_rh_Lat_Fis-ant-Horizont',
        'wm_rh_Lat_Fis-ant-Vertical',
        'wm_rh_Lat_Fis-post',
        'wm_rh_Medial_wall',
        'wm_rh_Pole_occipital',
        'wm_rh_Pole_temporal',
        'wm_rh_S_calcarine',
        'wm_rh_S_central',
        'wm_rh_S_cingul-Marginalis',
        'wm_rh_S_circular_insula_ant',
        'wm_rh_S_circular_insula_inf',
        'wm_rh_S_circular_insula_sup',
        'wm_rh_S_collat_transv_ant',
        'wm_rh_S_collat_transv_post',
        'wm_rh_S_front_inf',
        'wm_rh_S_front_middle',
        'wm_rh_S_front_sup',
        'wm_rh_S_interm_prim-Jensen',
        'wm_rh_S_intrapariet_and_P_trans',
        'wm_rh_S_oc_middle_and_Lunatus',
        'wm_rh_S_oc_sup_and_transversal',
        'wm_rh_S_occipital_ant',
        'wm_rh_S_oc-temp_lat',
        'wm_rh_S_oc-temp_med_and_Lingual',
        'wm_rh_S_orbital_lateral',
        'wm_rh_S_orbital_med-olfact',
        'wm_rh_S_orbital-H_Shaped',
        'wm_rh_S_parieto_occipital',
        'wm_rh_S_pericallosal',
        'wm_rh_S_postcentral',
        'wm_rh_S_precentral-inf-part',
        'wm_rh_S_precentral-sup-part',
        'wm_rh_S_suborbital',
        'wm_rh_S_subparietal',
        'wm_rh_S_temporal_inf',
        'wm_rh_S_temporal_sup',
        'wm_rh_S_temporal_transverse']

    colors = [[0, 0, 0],
        [70, 130, 180],
        [245, 245, 245],
        [205, 62, 78],
        [120, 18, 134],
        [196, 58, 250],
        [0, 148, 0],
        [220, 248, 164],
        [230, 148, 34],
        [0, 118, 14],
        [0, 118, 14],
        [122, 186, 220],
        [236, 13, 176],
        [12, 48, 255],
        [204, 182, 142],
        [42, 204, 164],
        [119, 159, 176],
        [220, 216, 20],
        [103, 255, 255],
        [80, 196, 98],
        [60, 58, 210],
        [60, 58, 210],
        [60, 58, 210],
        [60, 58, 210],
        [60, 60, 60],
        [255, 165, 0],
        [255, 165, 0],
        [0, 255, 127],
        [165, 42, 42],
        [135, 206, 235],
        [160, 32, 240],
        [0, 200, 200],
        [100, 50, 100],
        [135, 50, 74],
        [122, 135, 50],
        [51, 50, 135],
        [74, 155, 60],
        [120, 62, 43],
        [74, 155, 60],
        [122, 135, 50],
        [70, 130, 180],
        [0, 225, 0],
        [205, 62, 78],
        [120, 18, 134],
        [196, 58, 250],
        [0, 148, 0],
        [220, 248, 164],
        [230, 148, 34],
        [0, 118, 14],
        [0, 118, 14],
        [122, 186, 220],
        [236, 13, 176],
        [13, 48, 255],
        [220, 216, 20],
        [103, 255, 255],
        [80, 196, 98],
        [60, 58, 210],
        [255, 165, 0],
        [255, 165, 0],
        [0, 255, 127],
        [165, 42, 42],
        [135, 206, 235],
        [160, 32, 240],
        [0, 200, 221],
        [100, 50, 100],
        [135, 50, 74],
        [122, 135, 50],
        [51, 50, 135],
        [74, 155, 60],
        [120, 62, 43],
        [74, 155, 60],
        [122, 135, 50],
        [120, 190, 150],
        [122, 135, 50],
        [122, 135, 50],
        [200, 70, 255],
        [255, 148, 10],
        [255, 148, 10],
        [164, 108, 226],
        [164, 108, 226],
        [164, 108, 226],
        [255, 218, 185],
        [255, 218, 185],
        [234, 169, 30],
        [250, 255, 50],
        [200, 120, 255],
        [200, 121, 255],
        [200, 122, 255],
        [205, 10, 125],
        [205, 10, 125],
        [160, 32, 240],
        [124, 140, 178],
        [125, 140, 178],
        [126, 140, 178],
        [127, 140, 178],
        [124, 141, 178],
        [124, 142, 178],
        [124, 143, 178],
        [124, 144, 178],
        [124, 140, 179],
        [124, 140, 178],
        [125, 140, 178],
        [126, 140, 178],
        [127, 140, 178],
        [124, 141, 178],
        [124, 142, 178],
        [124, 143, 178],
        [124, 144, 178],
        [124, 140, 179],
        [255, 20, 147],
        [205, 179, 139],
        [238, 238, 209],
        [200, 200, 200],
        [74, 255, 74],
        [238, 0, 0],
        [0, 0, 139],
        [173, 255, 47],
        [133, 203, 229],
        [26, 237, 57],
        [34, 139, 34],
        [30, 144, 255],
        [147, 19, 173],
        [238, 59, 59],
        [221, 39, 200],
        [238, 174, 238],
        [255, 0, 0],
        [72, 61, 139],
        [21, 39, 132],
        [21, 39, 132],
        [65, 135, 20],
        [65, 135, 20],
        [134, 4, 160],
        [221, 226, 68],
        [255, 255, 254],
        [52, 209, 226],
        [239, 160, 223],
        [70, 130, 180],
        [70, 130, 181],
        [139, 121, 94],
        [224, 224, 224],
        [255, 0, 0],
        [205, 205, 0],
        [238, 238, 209],
        [139, 121, 94],
        [238, 59, 59],
        [238, 59, 59],
        [238, 59, 59],
        [62, 10, 205],
        [62, 10, 205],
        [0, 118, 14],
        [0, 118, 14],
        [220, 216, 21],
        [220, 216, 21],
        [122, 186, 220],
        [122, 186, 220],
        [255, 165, 0],
        [14, 48, 255],
        [166, 42, 42],
        [121, 18, 134],
        [236, 13, 127],
        [236, 13, 126],
        [119, 159, 176],
        [119, 0, 176],
        [119, 100, 176],
        [119, 200, 176],
        [119, 159, 100],
        [119, 159, 200],
        [73, 61, 139],
        [73, 62, 139],
        [0, 196, 255],
        [255, 164, 164],
        [196, 196, 0],
        [0, 100, 255],
        [128, 196, 164],
        [0, 126, 75],
        [128, 96, 64],
        [0, 50, 128],
        [255, 204, 153],
        [255, 128, 128],
        [255, 255, 0],
        [64, 0, 64],
        [0, 0, 255],
        [255, 0, 0],
        [128, 128, 255],
        [0, 128, 0],
        [196, 160, 128],
        [32, 200, 255],
        [128, 255, 128],
        [204, 153, 204],
        [121, 17, 136],
        [128, 0, 0],
        [128, 32, 255],
        [255, 204, 102],
        [128, 128, 128],
        [104, 255, 255],
        [0, 226, 0],
        [205, 63, 78],
        [197, 58, 250],
        [33, 150, 250],
        [226, 0, 0],
        [100, 100, 100],
        [197, 150, 250],
        [170, 170, 255],
        [255, 0, 0],
        [0, 0, 64],
        [0, 0, 112],
        [0, 0, 160],
        [0, 0, 208],
        [0, 0, 255],
        [0, 0, 0],
        [255, 0, 0],
        [255, 80, 0],
        [255, 160, 0],
        [255, 255, 0],
        [0, 255, 0],
        [255, 0, 160],
        [255, 0, 255],
        [255, 50, 80],
        [80, 255, 50],
        [160, 255, 50],
        [160, 200, 255],
        [0, 255, 160],
        [0, 0, 255],
        [80, 50, 255],
        [160, 0, 255],
        [255, 210, 0],
        [0, 160, 255],
        [255, 200, 80],
        [255, 200, 160],
        [255, 80, 200],
        [255, 160, 200],
        [30, 255, 80],
        [80, 200, 255],
        [80, 255, 200],
        [195, 255, 200],
        [120, 200, 20],
        [170, 10, 200],
        [20, 130, 180],
        [20, 180, 130],
        [206, 62, 78],
        [121, 18, 134],
        [199, 58, 250],
        [1, 148, 0],
        [221, 248, 164],
        [231, 148, 34],
        [1, 118, 14],
        [120, 118, 14],
        [123, 186, 221],
        [238, 13, 177],
        [123, 186, 220],
        [138, 13, 206],
        [238, 130, 176],
        [218, 230, 76],
        [38, 213, 176],
        [1, 225, 176],
        [1, 225, 176],
        [200, 2, 100],
        [200, 2, 100],
        [5, 200, 90],
        [5, 200, 90],
        [100, 5, 200],
        [25, 255, 100],
        [25, 255, 100],
        [230, 7, 100],
        [230, 7, 100],
        [100, 5, 200],
        [150, 10, 200],
        [150, 10, 200],
        [175, 10, 176],
        [175, 10, 176],
        [10, 100, 255],
        [10, 100, 255],
        [150, 45, 70],
        [150, 45, 70],
        [45, 200, 15],
        [45, 200, 15],
        [227, 45, 100],
        [227, 45, 100],
        [227, 45, 100],
        [143, 188, 143],
        [255, 248, 220],
        [17, 85, 136],
        [119, 187, 102],
        [204, 68, 34],
        [204, 0, 255],
        [221, 187, 17],
        [153, 221, 238],
        [51, 17, 17],
        [0, 119, 85],
        [20, 100, 200],
        [17, 85, 137],
        [119, 187, 103],
        [204, 68, 35],
        [204, 0, 254],
        [221, 187, 16],
        [153, 221, 239],
        [51, 17, 18],
        [0, 119, 86],
        [20, 100, 201],
        [254, 254, 254],
        [70, 130, 180],
        [245, 245, 245],
        [205, 62, 78],
        [120, 18, 134],
        [196, 58, 250],
        [0, 148, 0],
        [220, 248, 164],
        [230, 148, 34],
        [0, 118, 14],
        [0, 118, 14],
        [122, 186, 220],
        [236, 13, 176],
        [12, 48, 255],
        [204, 182, 142],
        [42, 204, 164],
        [119, 159, 176],
        [220, 216, 20],
        [103, 255, 255],
        [80, 196, 98],
        [60, 58, 210],
        [60, 58, 210],
        [60, 58, 210],
        [60, 58, 210],
        [60, 60, 60],
        [255, 165, 0],
        [255, 165, 0],
        [0, 255, 127],
        [165, 42, 42],
        [204, 0, 0],
        [255, 0, 0],
        [0, 0, 255],
        [30, 144, 255],
        [100, 212, 237],
        [218, 165, 32],
        [255, 215, 0],
        [255, 255, 166],
        [153, 0, 204],
        [153, 141, 209],
        [204, 204, 255],
        [31, 212, 194],
        [3, 255, 237],
        [204, 255, 255],
        [86, 74, 147],
        [114, 114, 190],
        [184, 178, 255],
        [126, 138, 37],
        [189, 197, 117],
        [240, 230, 140],
        [204, 0, 0],
        [255, 0, 0],
        [0, 0, 255],
        [30, 144, 255],
        [100, 212, 237],
        [218, 165, 32],
        [255, 215, 0],
        [255, 255, 166],
        [153, 0, 204],
        [153, 141, 209],
        [204, 204, 255],
        [31, 212, 194],
        [3, 255, 237],
        [204, 255, 255],
        [86, 74, 147],
        [114, 114, 190],
        [184, 178, 255],
        [126, 138, 37],
        [189, 197, 117],
        [240, 230, 140],
        [120, 18, 134],
        [205, 62, 78],
        [0, 225, 0],
        [255, 100, 100],
        [25, 5, 25],
        [25, 100, 40],
        [125, 100, 160],
        [100, 25, 0],
        [120, 70, 50],
        [220, 20, 100],
        [220, 20, 10],
        [180, 220, 140],
        [220, 60, 220],
        [180, 40, 120],
        [140, 20, 140],
        [20, 30, 140],
        [35, 75, 50],
        [225, 140, 140],
        [200, 35, 75],
        [160, 100, 50],
        [20, 220, 60],
        [60, 220, 60],
        [220, 180, 140],
        [20, 100, 50],
        [220, 60, 20],
        [120, 100, 60],
        [220, 20, 20],
        [220, 180, 220],
        [60, 20, 220],
        [160, 140, 180],
        [80, 20, 140],
        [75, 50, 125],
        [20, 220, 160],
        [20, 180, 140],
        [140, 220, 220],
        [80, 160, 20],
        [100, 0, 100],
        [70, 70, 70],
        [150, 150, 200],
        [255, 192, 32],
        [25, 5, 25],
        [25, 100, 40],
        [125, 100, 160],
        [100, 25, 0],
        [120, 70, 50],
        [220, 20, 100],
        [220, 20, 10],
        [180, 220, 140],
        [220, 60, 220],
        [180, 40, 120],
        [140, 20, 140],
        [20, 30, 140],
        [35, 75, 50],
        [225, 140, 140],
        [200, 35, 75],
        [160, 100, 50],
        [20, 220, 60],
        [60, 220, 60],
        [220, 180, 140],
        [20, 100, 50],
        [220, 60, 20],
        [120, 100, 60],
        [220, 20, 20],
        [220, 180, 220],
        [60, 20, 220],
        [160, 140, 180],
        [80, 20, 140],
        [75, 50, 125],
        [20, 220, 160],
        [20, 180, 140],
        [140, 220, 220],
        [80, 160, 20],
        [100, 0, 100],
        [70, 70, 70],
        [150, 150, 200],
        [255, 192, 32],
        [230, 250, 230],
        [230, 155, 215],
        [130, 155, 95],
        [155, 230, 255],
        [135, 185, 205],
        [35, 235, 155],
        [35, 235, 245],
        [75, 35, 115],
        [35, 195, 35],
        [75, 215, 135],
        [115, 235, 115],
        [235, 225, 115],
        [220, 180, 205],
        [30, 115, 115],
        [55, 220, 180],
        [95, 155, 205],
        [235, 35, 195],
        [195, 35, 195],
        [35, 75, 115],
        [235, 155, 205],
        [35, 195, 235],
        [135, 155, 195],
        [35, 235, 235],
        [35, 75, 35],
        [195, 235, 35],
        [95, 115, 75],
        [175, 235, 115],
        [180, 205, 130],
        [235, 35, 95],
        [235, 75, 115],
        [115, 35, 35],
        [175, 95, 235],
        [155, 255, 155],
        [185, 185, 185],
        [105, 105, 55],
        [254, 191, 31],
        [230, 250, 230],
        [230, 155, 215],
        [130, 155, 95],
        [155, 230, 255],
        [135, 185, 205],
        [35, 235, 155],
        [35, 235, 245],
        [75, 35, 115],
        [35, 195, 35],
        [75, 215, 135],
        [115, 235, 115],
        [235, 225, 115],
        [220, 180, 205],
        [30, 115, 115],
        [55, 220, 180],
        [95, 155, 205],
        [235, 35, 195],
        [195, 35, 195],
        [35, 75, 115],
        [235, 155, 205],
        [35, 195, 235],
        [135, 155, 195],
        [35, 235, 235],
        [35, 75, 35],
        [195, 235, 35],
        [95, 115, 75],
        [175, 235, 115],
        [180, 205, 130],
        [235, 35, 95],
        [235, 75, 115],
        [115, 35, 35],
        [175, 95, 235],
        [155, 255, 155],
        [185, 185, 185],
        [105, 105, 55],
        [254, 191, 31],
        [0, 0, 0],
        [50, 50, 50],
        [180, 20, 30],
        [60, 25, 25],
        [25, 60, 60],
        [25, 60, 61],
        [25, 90, 60],
        [25, 120, 60],
        [25, 150, 60],
        [25, 180, 60],
        [25, 210, 60],
        [25, 150, 90],
        [25, 180, 90],
        [25, 210, 90],
        [180, 20, 20],
        [220, 20, 100],
        [140, 60, 60],
        [180, 220, 140],
        [140, 100, 180],
        [180, 20, 140],
        [140, 20, 140],
        [21, 10, 10],
        [225, 140, 140],
        [23, 60, 180],
        [180, 60, 180],
        [20, 220, 60],
        [60, 20, 140],
        [220, 180, 140],
        [65, 100, 20],
        [220, 60, 20],
        [60, 100, 60],
        [20, 60, 220],
        [100, 100, 60],
        [220, 180, 220],
        [20, 180, 140],
        [60, 140, 180],
        [25, 20, 140],
        [20, 60, 100],
        [60, 220, 20],
        [60, 20, 220],
        [220, 220, 100],
        [180, 60, 60],
        [60, 60, 220],
        [220, 60, 220],
        [65, 220, 60],
        [25, 140, 20],
        [13, 0, 250],
        [61, 20, 220],
        [61, 20, 60],
        [61, 60, 100],
        [25, 25, 25],
        [140, 20, 60],
        [220, 180, 20],
        [63, 180, 180],
        [221, 20, 10],
        [21, 220, 20],
        [183, 100, 20],
        [221, 20, 100],
        [221, 60, 140],
        [221, 20, 220],
        [61, 220, 220],
        [100, 200, 200],
        [10, 200, 200],
        [221, 220, 20],
        [141, 20, 100],
        [61, 220, 100],
        [21, 220, 60],
        [141, 60, 20],
        [143, 20, 220],
        [61, 20, 180],
        [101, 60, 220],
        [21, 20, 140],
        [221, 140, 20],
        [141, 100, 220],
        [101, 20, 20],
        [221, 100, 20],
        [181, 200, 20],
        [21, 180, 140],
        [101, 100, 180],
        [181, 220, 20],
        [21, 140, 200],
        [21, 20, 240],
        [21, 20, 200],
        [61, 180, 60],
        [61, 180, 250],
        [21, 20, 60],
        [101, 60, 60],
        [21, 220, 220],
        [21, 180, 180],
        [223, 220, 60],
        [221, 60, 60],
        [0, 0, 0],
        [50, 50, 50],
        [180, 20, 30],
        [60, 25, 25],
        [25, 60, 60],
        [180, 20, 20],
        [220, 20, 100],
        [140, 60, 60],
        [180, 220, 140],
        [140, 100, 180],
        [180, 20, 140],
        [140, 20, 140],
        [21, 10, 10],
        [225, 140, 140],
        [23, 60, 180],
        [180, 60, 180],
        [20, 220, 60],
        [60, 20, 140],
        [220, 180, 140],
        [65, 100, 20],
        [220, 60, 20],
        [60, 100, 60],
        [20, 60, 220],
        [100, 100, 60],
        [220, 180, 220],
        [20, 180, 140],
        [60, 140, 180],
        [25, 20, 140],
        [20, 60, 100],
        [60, 220, 20],
        [60, 20, 220],
        [220, 220, 100],
        [180, 60, 60],
        [60, 60, 220],
        [220, 60, 220],
        [65, 220, 60],
        [25, 140, 20],
        [13, 0, 250],
        [61, 20, 220],
        [61, 20, 60],
        [61, 60, 100],
        [25, 25, 25],
        [140, 20, 60],
        [220, 180, 20],
        [63, 180, 180],
        [221, 20, 10],
        [21, 220, 20],
        [183, 100, 20],
        [221, 20, 100],
        [221, 60, 140],
        [221, 20, 220],
        [61, 220, 220],
        [100, 200, 200],
        [10, 200, 200],
        [221, 220, 20],
        [141, 20, 100],
        [61, 220, 100],
        [21, 220, 60],
        [141, 60, 20],
        [143, 20, 220],
        [61, 20, 180],
        [101, 60, 220],
        [21, 20, 140],
        [221, 140, 20],
        [141, 100, 220],
        [101, 20, 20],
        [221, 100, 20],
        [181, 200, 20],
        [21, 180, 140],
        [101, 100, 180],
        [181, 220, 20],
        [21, 140, 200],
        [21, 20, 240],
        [21, 20, 200],
        [61, 180, 60],
        [61, 180, 250],
        [21, 20, 60],
        [101, 60, 60],
        [21, 220, 220],
        [21, 180, 180],
        [223, 220, 60],
        [221, 60, 60],
        [25, 60, 61],
        [25, 90, 60],
        [25, 120, 60],
        [25, 150, 60],
        [25, 180, 60],
        [25, 210, 60],
        [25, 150, 90],
        [25, 180, 90],
        [25, 210, 90],
        [0, 0, 0],
        [50, 50, 50],
        [180, 20, 30],
        [60, 25, 25],
        [25, 60, 60],
        [180, 20, 20],
        [220, 20, 100],
        [140, 60, 60],
        [180, 220, 140],
        [140, 100, 180],
        [180, 20, 140],
        [140, 20, 140],
        [21, 10, 10],
        [225, 140, 140],
        [23, 60, 180],
        [180, 60, 180],
        [20, 220, 60],
        [60, 20, 140],
        [220, 180, 140],
        [65, 100, 20],
        [220, 60, 20],
        [60, 100, 60],
        [20, 60, 220],
        [100, 100, 60],
        [220, 180, 220],
        [20, 180, 140],
        [60, 140, 180],
        [25, 20, 140],
        [20, 60, 100],
        [60, 220, 20],
        [60, 20, 220],
        [220, 220, 100],
        [180, 60, 60],
        [60, 60, 220],
        [220, 60, 220],
        [65, 220, 60],
        [25, 140, 20],
        [13, 0, 250],
        [61, 20, 220],
        [61, 20, 60],
        [61, 60, 100],
        [25, 25, 25],
        [140, 20, 60],
        [220, 180, 20],
        [63, 180, 180],
        [221, 20, 10],
        [21, 220, 20],
        [183, 100, 20],
        [221, 20, 100],
        [221, 60, 140],
        [221, 20, 220],
        [61, 220, 220],
        [100, 200, 200],
        [10, 200, 200],
        [221, 220, 20],
        [141, 20, 100],
        [61, 220, 100],
        [21, 220, 60],
        [141, 60, 20],
        [143, 20, 220],
        [61, 20, 180],
        [101, 60, 220],
        [21, 20, 140],
        [221, 140, 20],
        [141, 100, 220],
        [101, 20, 20],
        [221, 100, 20],
        [181, 200, 20],
        [21, 180, 140],
        [101, 100, 180],
        [181, 220, 20],
        [21, 140, 200],
        [21, 20, 240],
        [21, 20, 200],
        [61, 180, 60],
        [61, 180, 250],
        [21, 20, 60],
        [101, 60, 60],
        [21, 220, 220],
        [21, 180, 180],
        [223, 220, 60],
        [221, 60, 60],
        [0, 0, 0],
        [50, 50, 50],
        [180, 20, 30],
        [60, 25, 25],
        [25, 60, 60],
        [180, 20, 20],
        [220, 20, 100],
        [140, 60, 60],
        [180, 220, 140],
        [140, 100, 180],
        [180, 20, 140],
        [140, 20, 140],
        [21, 10, 10],
        [225, 140, 140],
        [23, 60, 180],
        [180, 60, 180],
        [20, 220, 60],
        [60, 20, 140],
        [220, 180, 140],
        [65, 100, 20],
        [220, 60, 20],
        [60, 100, 60],
        [20, 60, 220],
        [100, 100, 60],
        [220, 180, 220],
        [20, 180, 140],
        [60, 140, 180],
        [25, 20, 140],
        [20, 60, 100],
        [60, 220, 20],
        [60, 20, 220],
        [220, 220, 100],
        [180, 60, 60],
        [60, 60, 220],
        [220, 60, 220],
        [65, 220, 60],
        [25, 140, 20],
        [13, 0, 250],
        [61, 20, 220],
        [61, 20, 60],
        [61, 60, 100],
        [25, 25, 25],
        [140, 20, 60],
        [220, 180, 20],
        [63, 180, 180],
        [221, 20, 10],
        [21, 220, 20],
        [183, 100, 20],
        [221, 20, 100],
        [221, 60, 140],
        [221, 20, 220],
        [61, 220, 220],
        [100, 200, 200],
        [10, 200, 200],
        [221, 220, 20],
        [141, 20, 100],
        [61, 220, 100],
        [21, 220, 60],
        [141, 60, 20],
        [143, 20, 220],
        [61, 20, 180],
        [101, 60, 220],
        [21, 20, 140],
        [221, 140, 20],
        [141, 100, 220],
        [101, 20, 20],
        [221, 100, 20],
        [181, 200, 20],
        [21, 180, 140],
        [101, 100, 180],
        [181, 220, 20],
        [21, 140, 200],
        [21, 20, 240],
        [21, 20, 200],
        [61, 180, 60],
        [61, 180, 250],
        [21, 20, 60],
        [101, 60, 60],
        [21, 220, 220],
        [21, 180, 180],
        [223, 220, 60],
        [221, 60, 60],
        [20, 30, 40],
        [20, 30, 40],
        [204, 102, 102],
        [204, 102, 102],
        [255, 255, 102],
        [153, 204, 0],
        [0, 153, 153],
        [204, 153, 255],
        [255, 153, 51],
        [204, 204, 204],
        [153, 255, 255],
        [102, 153, 255],
        [255, 255, 102],
        [153, 204, 0],
        [0, 153, 153],
        [204, 153, 255],
        [255, 153, 51],
        [204, 204, 204],
        [153, 255, 255],
        [102, 153, 255],
        [204, 102, 102],
        [204, 102, 102],
        [255, 255, 102],
        [153, 204, 0],
        [0, 153, 153],
        [204, 153, 255],
        [255, 153, 51],
        [204, 204, 204],
        [153, 255, 255],
        [102, 153, 255],
        [255, 255, 102],
        [153, 204, 0],
        [0, 153, 153],
        [204, 153, 255],
        [255, 153, 51],
        [204, 204, 204],
        [153, 255, 255],
        [102, 153, 255],
        [0, 255, 0],
        [255, 255, 0],
        [0, 255, 255],
        [0, 0, 255],
        [236, 16, 231],
        [237, 18, 232],
        [236, 13, 227],
        [236, 17, 228],
        [1, 255, 1],
        [2, 255, 1],
        [236, 14, 230],
        [237, 14, 230],
        [72, 132, 181],
        [243, 243, 243],
        [207, 63, 79],
        [121, 20, 135],
        [197, 60, 248],
        [2, 149, 2],
        [221, 249, 166],
        [232, 146, 35],
        [20, 60, 120],
        [250, 250, 0],
        [122, 187, 222],
        [237, 12, 177],
        [10, 49, 255],
        [205, 184, 144],
        [45, 205, 165],
        [117, 160, 175],
        [221, 217, 21],
        [20, 60, 120],
        [141, 21, 100],
        [225, 140, 141],
        [42, 201, 168],
        [168, 104, 162],
        [74, 130, 181],
        [242, 241, 240],
        [206, 65, 78],
        [120, 21, 133],
        [195, 61, 246],
        [3, 147, 6],
        [220, 251, 163],
        [232, 146, 33],
        [4, 114, 14],
        [121, 184, 220],
        [235, 11, 175],
        [12, 46, 250],
        [203, 182, 143],
        [42, 204, 167],
        [30, 5, 30],
        [30, 100, 45],
        [130, 100, 165],
        [105, 25, 5],
        [125, 70, 55],
        [225, 20, 105],
        [225, 20, 15],
        [30, 55, 30],
        [30, 150, 45],
        [130, 150, 165],
        [105, 75, 5],
        [125, 120, 55],
        [225, 70, 105],
        [225, 70, 15],
        [0, 0, 0],
        [23, 220, 60],
        [23, 60, 180],
        [63, 100, 60],
        [63, 20, 220],
        [13, 0, 250],
        [26, 60, 0],
        [26, 60, 75],
        [26, 60, 150],
        [25, 60, 250],
        [60, 25, 25],
        [180, 20, 20],
        [220, 20, 100],
        [140, 60, 60],
        [180, 220, 140],
        [140, 100, 180],
        [180, 20, 140],
        [23, 10, 10],
        [225, 140, 140],
        [180, 60, 180],
        [20, 220, 60],
        [60, 20, 140],
        [220, 180, 140],
        [65, 100, 20],
        [220, 60, 20],
        [20, 60, 220],
        [100, 100, 60],
        [220, 180, 220],
        [20, 180, 140],
        [60, 140, 180],
        [25, 20, 140],
        [20, 60, 100],
        [60, 220, 20],
        [60, 60, 220],
        [220, 60, 220],
        [65, 220, 60],
        [25, 140, 20],
        [220, 220, 100],
        [180, 60, 60],
        [61, 20, 220],
        [61, 20, 60],
        [61, 60, 100],
        [25, 25, 25],
        [140, 20, 60],
        [220, 180, 20],
        [63, 180, 180],
        [221, 20, 10],
        [221, 20, 100],
        [221, 60, 140],
        [221, 20, 220],
        [61, 220, 220],
        [100, 200, 200],
        [10, 200, 200],
        [221, 220, 20],
        [141, 20, 100],
        [61, 220, 100],
        [141, 60, 20],
        [143, 20, 220],
        [101, 60, 220],
        [21, 20, 140],
        [61, 20, 180],
        [221, 140, 20],
        [141, 100, 220],
        [221, 100, 20],
        [181, 200, 20],
        [101, 20, 20],
        [101, 100, 180],
        [181, 220, 20],
        [21, 140, 200],
        [21, 20, 240],
        [21, 20, 200],
        [21, 20, 60],
        [101, 60, 60],
        [21, 180, 180],
        [223, 220, 60],
        [221, 60, 60],
        [0, 0, 0],
        [23, 220, 60],
        [23, 60, 180],
        [63, 100, 60],
        [63, 20, 220],
        [13, 0, 250],
        [26, 60, 0],
        [26, 60, 75],
        [26, 60, 150],
        [25, 60, 250],
        [60, 25, 25],
        [180, 20, 20],
        [220, 20, 100],
        [140, 60, 60],
        [180, 220, 140],
        [140, 100, 180],
        [180, 20, 140],
        [23, 10, 10],
        [225, 140, 140],
        [180, 60, 180],
        [20, 220, 60],
        [60, 20, 140],
        [220, 180, 140],
        [65, 100, 20],
        [220, 60, 20],
        [20, 60, 220],
        [100, 100, 60],
        [220, 180, 220],
        [20, 180, 140],
        [60, 140, 180],
        [25, 20, 140],
        [20, 60, 100],
        [60, 220, 20],
        [60, 60, 220],
        [220, 60, 220],
        [65, 220, 60],
        [25, 140, 20],
        [220, 220, 100],
        [180, 60, 60],
        [61, 20, 220],
        [61, 20, 60],
        [61, 60, 100],
        [25, 25, 25],
        [140, 20, 60],
        [220, 180, 20],
        [63, 180, 180],
        [221, 20, 10],
        [221, 20, 100],
        [221, 60, 140],
        [221, 20, 220],
        [61, 220, 220],
        [100, 200, 200],
        [10, 200, 200],
        [221, 220, 20],
        [141, 20, 100],
        [61, 220, 100],
        [141, 60, 20],
        [143, 20, 220],
        [101, 60, 220],
        [21, 20, 140],
        [61, 20, 180],
        [221, 140, 20],
        [141, 100, 220],
        [221, 100, 20],
        [181, 200, 20],
        [101, 20, 20],
        [101, 100, 180],
        [181, 220, 20],
        [21, 140, 200],
        [21, 20, 240],
        [21, 20, 200],
        [21, 20, 60],
        [101, 60, 60],
        [21, 180, 180],
        [223, 220, 60],
        [221, 60, 60],
        [0, 0, 0],
        [23, 220, 60],
        [23, 60, 180],
        [63, 100, 60],
        [63, 20, 220],
        [13, 0, 250],
        [26, 60, 0],
        [26, 60, 75],
        [26, 60, 150],
        [25, 60, 250],
        [60, 25, 25],
        [180, 20, 20],
        [220, 20, 100],
        [140, 60, 60],
        [180, 220, 140],
        [140, 100, 180],
        [180, 20, 140],
        [23, 10, 10],
        [225, 140, 140],
        [180, 60, 180],
        [20, 220, 60],
        [60, 20, 140],
        [220, 180, 140],
        [65, 100, 20],
        [220, 60, 20],
        [20, 60, 220],
        [100, 100, 60],
        [220, 180, 220],
        [20, 180, 140],
        [60, 140, 180],
        [25, 20, 140],
        [20, 60, 100],
        [60, 220, 20],
        [60, 60, 220],
        [220, 60, 220],
        [65, 220, 60],
        [25, 140, 20],
        [220, 220, 100],
        [180, 60, 60],
        [61, 20, 220],
        [61, 20, 60],
        [61, 60, 100],
        [25, 25, 25],
        [140, 20, 60],
        [220, 180, 20],
        [63, 180, 180],
        [221, 20, 10],
        [221, 20, 100],
        [221, 60, 140],
        [221, 20, 220],
        [61, 220, 220],
        [100, 200, 200],
        [10, 200, 200],
        [221, 220, 20],
        [141, 20, 100],
        [61, 220, 100],
        [141, 60, 20],
        [143, 20, 220],
        [101, 60, 220],
        [21, 20, 140],
        [61, 20, 180],
        [221, 140, 20],
        [141, 100, 220],
        [221, 100, 20],
        [181, 200, 20],
        [101, 20, 20],
        [101, 100, 180],
        [181, 220, 20],
        [21, 140, 200],
        [21, 20, 240],
        [21, 20, 200],
        [21, 20, 60],
        [101, 60, 60],
        [21, 180, 180],
        [223, 220, 60],
        [221, 60, 60],
        [0, 0, 0],
        [23, 220, 60],
        [23, 60, 180],
        [63, 100, 60],
        [63, 20, 220],
        [13, 0, 250],
        [26, 60, 0],
        [26, 60, 75],
        [26, 60, 150],
        [25, 60, 250],
        [60, 25, 25],
        [180, 20, 20],
        [220, 20, 100],
        [140, 60, 60],
        [180, 220, 140],
        [140, 100, 180],
        [180, 20, 140],
        [23, 10, 10],
        [225, 140, 140],
        [180, 60, 180],
        [20, 220, 60],
        [60, 20, 140],
        [220, 180, 140],
        [65, 100, 20],
        [220, 60, 20],
        [20, 60, 220],
        [100, 100, 60],
        [220, 180, 220],
        [20, 180, 140],
        [60, 140, 180],
        [25, 20, 140],
        [20, 60, 100],
        [60, 220, 20],
        [60, 60, 220],
        [220, 60, 220],
        [65, 220, 60],
        [25, 140, 20],
        [220, 220, 100],
        [180, 60, 60],
        [61, 20, 220],
        [61, 20, 60],
        [61, 60, 100],
        [25, 25, 25],
        [140, 20, 60],
        [220, 180, 20],
        [63, 180, 180],
        [221, 20, 10],
        [221, 20, 100],
        [221, 60, 140],
        [221, 20, 220],
        [61, 220, 220],
        [100, 200, 200],
        [10, 200, 200],
        [221, 220, 20],
        [141, 20, 100],
        [61, 220, 100],
        [141, 60, 20],
        [143, 20, 220],
        [101, 60, 220],
        [21, 20, 140],
        [61, 20, 180],
        [221, 140, 20],
        [141, 100, 220],
        [221, 100, 20],
        [181, 200, 20],
        [101, 20, 20],
        [101, 100, 180],
        [181, 220, 20],
        [21, 140, 200],
        [21, 20, 240],
        [21, 20, 200],
        [21, 20, 60],
        [101, 60, 60],
        [21, 180, 180],
        [223, 220, 60],
        [221, 60, 60]]


    return numbers, names, colors


class DKTprotocol(object):
    """Variables related to the Desikan-Killiany-Tourville labeling protocol.

    For more information about the Desikan-Killiany-Tourville (DKT) human
    brain cortical labeling protocol, see http://mindboggle.info/data/
    and the article:

    http://www.frontiersin.org/Brain_Imaging_Methods/10.3389/fnins.2012.00171/full
    "101 labeled brain images and a consistent human cortical labeling protocol"
    Arno Klein, Jason Tourville. Frontiers in Brain Imaging Methods. 6:171.
    DOI: 10.3389/fnins.2012.00171

    Returns
    -------
    [left, right]_cerebrum_cortex_[DKT31_][numbers, names, colors]
    [left, right]_ventricle_[numbers, names, colors]
    medial_ventricle_[numbers, names, colors]
    [left, right]_cerebrum_noncortex_[numbers, names, colors]
    medial_cerebrum_noncortex_[numbers, names, colors]
    [left, right]_cerebellum_cortex_[numbers, names, colors]
    [left, right]_cerebellum_noncortex_[numbers, names, colors]
    medial_cerebellum_noncortex_[numbers, names, colors]
    brainstem_[numbers, names, colors]
    extra_[numbers, names, colors]
    misc_[numbers, names, colors]
    ventricle_[numbers, names, colors]
    cerebrum_cortex_[numbers, names, colors]
    cerebrum_noncortex_[numbers, names, colors]
    [left, right]_cerebrum_[numbers, names, colors]
    cerebrum_[numbers, names, colors]
    [left, right]_cerebellum_[numbers, names, colors]
    cerebellum_cortex_[numbers, names, colors]
    cerebellum_noncortex_[numbers, names, colors]
    cerebellum_[numbers, names, colors]
    label_[numbers, names, colors]
    colormap : list of lists
    colormap_normalized : list of lists
    sulcus_[names[_abbr], numbers]
    unique_sulcus_label_pairs : list of unique pairs of integers
        unique label pairs corresponding to label boundaries / sulcus / fundus
    [left_, right_]sulcus_label_pair_lists : list of two lists of lists of integer pairs
        list containing left and/or right lists, each with multiple lists of
        integer pairs corresponding to label boundaries / sulcus / fundus

    Examples
    --------
    >>> from mindboggle.mio.labels import DKTprotocol
    >>> dkt = DKTprotocol()
    >>> dkt.left_cerebrum_names[0:3]
    ['Left-Cerebral-Cortex', 'Left-Insula', 'Left-Operculum']
    >>> dkt.left_cerebrum_numbers[0:10]
    [3, 19, 20, 1000, 1001, 1002, 1003, 1005, 1006, 1007]
    >>> dkt.left_cerebrum_colors[0]
    [205, 62, 78]

    """
    from mindboggle.mio.labels import return_numbers_names_colors

    # ------------------------------------------------------------------------
    # Return numbers, names, colors extracted from FreeSurferColorLUT.txt:
    # ------------------------------------------------------------------------
    numbers, names, colors = return_numbers_names_colors()

    # ------------------------------------------------------------------------
    # Cerebral cortex label ID numbers, names, groups (DKT31 protocol):
    # ------------------------------------------------------------------------
    DKT31_numbers = [2, 3] + list(range(5, 32)) + [34, 35]
    DKT31_names = ['caudal anterior cingulate',
                   'caudal middle frontal',
                   'cuneus',
                   'entorhinal',
                   'fusiform',
                   'inferior parietal',
                   'inferior temporal',
                   'isthmus cingulate',
                   'lateral occipital',
                   'lateral orbitofrontal',
                   'lingual',
                   'medial orbitofrontal',
                   'middle temporal',
                   'parahippocampal',
                   'paracentral',
                   'pars opercularis',
                   'pars orbitalis',
                   'pars triangularis',
                   'pericalcarine',
                   'postcentral',
                   'posterior cingulate',
                   'precentral',
                   'precuneus',
                   'rostral anterior cingulate',
                   'rostral middle frontal',
                   'superior frontal',
                   'superior parietal',
                   'superior temporal',
                   'supramarginal',
                   'transverse temporal',
                   'insula']
    # 1: frontal; 2: paracentral; 3: parietal;
    # 4: occipital; 5: temporal; 6: limbic
    DKT31_groups = [6, 1, 4, 6, 3, 5, 3, 6, 4, 1,
                    4, 1, 3, 6, 2, 1, 1, 1, 5, 5,
                    6, 1, 5, 6, 1, 1, 5, 3, 5, 3, 1]
    left_cerebrum_cortex_DKT31_list = [1000 + x for x in DKT31_numbers]
    right_cerebrum_cortex_DKT31_list = [2000 + x for x in DKT31_numbers]

    # ------------------------------------------------------------------------
    # Cerebral cortex label numbers (31 + duplicates):
    # ------------------------------------------------------------------------
    left_cerebrum_cortex_list = left_cerebrum_cortex_DKT31_list + \
        [3, 19, 20, 1000, 1001, 1032, 1033]
    right_cerebrum_cortex_list = right_cerebrum_cortex_DKT31_list + \
        [42, 55, 56, 2000, 2001, 2032, 2033]

    # ------------------------------------------------------------------------
    # Cerebral ventricle label numbers:
    # ------------------------------------------------------------------------
    left_ventricle_list = [4, 5]
    right_ventricle_list = [43, 44]
    medial_ventricle_list = [14, 15, 72]

    # ------------------------------------------------------------------------
    # Cerebral noncortex label numbers (including ventricles above):
    # ------------------------------------------------------------------------
    # These labels were converted from Neuromorphometrics BrainCOLOR subcortex
    # labels to be consistent with FreeSurferColorLUT.txt labels.
    #
    # Two labels did not have counterparts in FreeSurfer:
    #         [75, "left basal forebrain"],
    #         [76, "right basal forebrain"]
    # and were reassigned to unused numbers in FreeSurferColorLUT.txt:
    #         [91, "left basal forebrain"],
    #         [92, "right basal forebrain"]
    # ------------------------------------------------------------------------
    left_cerebrum_noncortex_list = \
        [2, 9, 10, 11, 12, 13, 17, 18, 25, 26, 27, 28, 30, 31, 78, 91, 96] + \
        list(range(100, 109)) + [155, 157] + list(range(550, 559)) + [1004] + \
        list(range(3000, 3036)) + [5001] + left_ventricle_list
    right_cerebrum_noncortex_list = \
        [41, 48, 49, 50, 51, 52, 53, 54, 57, 58, 59, 60, 62, 63, 79, 92, 97] + \
        list(range(109, 118)) + [156, 158] + list(range(500, 509)) + [2004] + \
        list(range(4000, 4036)) + [5002] + right_ventricle_list
    medial_cerebrum_noncortex_list = medial_ventricle_list + \
                                        [192] + list(range(250, 256))

    # ------------------------------------------------------------------------
    # Cerebellar label numbers:
    # ------------------------------------------------------------------------
    # These labels [71, 72, 73] in Neuromorphometrics BrainCOLOR subcortex
    # did not have counterparts in FreeSurferColorLUT.txt, and were
    # reassigned to unused numbers in FreeSurferColorLUT.txt:
    #   [630, "cerebellar vermal lobules I-V"],
    #   [631, "cerebellar vermal lobules VI-VII"],
    #   [632, "cerebellar vermal lobules VIII-X"],
    # ------------------------------------------------------------------------
    left_cerebellum_cortex_list = [6, 8]
    right_cerebellum_cortex_list = [45, 47]
    left_cerebellum_noncortex_list = [7]
    right_cerebellum_noncortex_list = [46]
    medial_cerebellum_noncortex_list = [630, 631, 632]

    # ------------------------------------------------------------------------
    # Extra label numbers:
    # ------------------------------------------------------------------------
    #  [16, "Brain stem"],
    #  [24, "CSF"],
    #  [85, "optic chiasm"]]
    #  170-175: brain stem
    # ------------------------------------------------------------------------
    brainstem_list = [16] + list(range(170, 176))
    extra_list = [24, 85]

    # ------------------------------------------------------------------------
    # Label numbers:
    # ------------------------------------------------------------------------
    left_cerebrum_cortex_DKT31_numbers = []
    right_cerebrum_cortex_DKT31_numbers = []
    left_cerebrum_cortex_numbers = []
    right_cerebrum_cortex_numbers = []
    left_ventricle_numbers = []
    right_ventricle_numbers = []
    medial_ventricle_numbers = []
    left_cerebrum_noncortex_numbers = []
    right_cerebrum_noncortex_numbers = []
    medial_cerebrum_noncortex_numbers = []
    left_cerebellum_cortex_numbers = []
    right_cerebellum_cortex_numbers = []
    left_cerebellum_noncortex_numbers = []
    right_cerebellum_noncortex_numbers = []
    medial_cerebellum_noncortex_numbers = []
    brainstem_numbers = []
    extra_numbers = []
    misc_numbers = []
    # ------------------------------------------------------------------------
    # Names corresponding to label numbers:
    # ------------------------------------------------------------------------
    left_cerebrum_cortex_DKT31_names = []
    right_cerebrum_cortex_DKT31_names = []
    left_cerebrum_cortex_names = []
    right_cerebrum_cortex_names = []
    left_ventricle_names = []
    right_ventricle_names = []
    medial_ventricle_names = []
    left_cerebrum_noncortex_names = []
    right_cerebrum_noncortex_names = []
    medial_cerebrum_noncortex_names = []
    left_cerebellum_cortex_names = []
    right_cerebellum_cortex_names = []
    left_cerebellum_noncortex_names = []
    right_cerebellum_noncortex_names = []
    medial_cerebellum_noncortex_names = []
    brainstem_names = []
    extra_names = []
    misc_names = []
    # ------------------------------------------------------------------------
    # Colors corresponding to label numbers:
    # ------------------------------------------------------------------------
    left_cerebrum_cortex_DKT31_colors = []
    right_cerebrum_cortex_DKT31_colors = []
    left_cerebrum_cortex_colors = []
    right_cerebrum_cortex_colors = []
    left_ventricle_colors = []
    right_ventricle_colors = []
    medial_ventricle_colors = []
    left_cerebrum_noncortex_colors = []
    right_cerebrum_noncortex_colors = []
    medial_cerebrum_noncortex_colors = []
    left_cerebellum_cortex_colors = []
    right_cerebellum_cortex_colors = []
    left_cerebellum_noncortex_colors = []
    right_cerebellum_noncortex_colors = []
    medial_cerebellum_noncortex_colors = []
    brainstem_colors = []
    extra_colors = []
    misc_colors = []

    # ------------------------------------------------------------------------
    # Lists of numbers, names, and colors:
    # ------------------------------------------------------------------------
    for i, n in enumerate(numbers):

        if n in left_cerebrum_cortex_DKT31_list:
            left_cerebrum_cortex_DKT31_numbers.append(numbers[i])
            left_cerebrum_cortex_DKT31_names.append(names[i])
            left_cerebrum_cortex_DKT31_colors.append(colors[i])
        elif n in right_cerebrum_cortex_DKT31_list:
            right_cerebrum_cortex_DKT31_numbers.append(numbers[i])
            right_cerebrum_cortex_DKT31_names.append(names[i])
            right_cerebrum_cortex_DKT31_colors.append(colors[i])

        if n in left_cerebrum_cortex_list:
            left_cerebrum_cortex_numbers.append(numbers[i])
            left_cerebrum_cortex_names.append(names[i])
            left_cerebrum_cortex_colors.append(colors[i])
        elif n in right_cerebrum_cortex_list:
            right_cerebrum_cortex_numbers.append(numbers[i])
            right_cerebrum_cortex_names.append(names[i])
            right_cerebrum_cortex_colors.append(colors[i])
        elif n in left_ventricle_list:
            left_ventricle_numbers.append(numbers[i])
            left_cerebrum_noncortex_numbers.append(numbers[i])
            left_ventricle_names.append(names[i])
            left_cerebrum_noncortex_names.append(names[i])
            left_ventricle_colors.append(colors[i])
            left_cerebrum_noncortex_colors.append(colors[i])
        elif n in right_ventricle_list:
            right_ventricle_numbers.append(numbers[i])
            right_cerebrum_noncortex_numbers.append(numbers[i])
            right_ventricle_names.append(names[i])
            right_cerebrum_noncortex_names.append(names[i])
            right_ventricle_colors.append(colors[i])
            right_cerebrum_noncortex_colors.append(colors[i])
        elif n in medial_ventricle_list:
            medial_ventricle_numbers.append(numbers[i])
            medial_cerebrum_noncortex_numbers.append(numbers[i])
            medial_ventricle_names.append(names[i])
            medial_cerebrum_noncortex_names.append(names[i])
            medial_ventricle_colors.append(colors[i])
            medial_cerebrum_noncortex_colors.append(colors[i])
        elif n in left_cerebrum_noncortex_list:
            left_cerebrum_noncortex_numbers.append(numbers[i])
            left_cerebrum_noncortex_names.append(names[i])
            left_cerebrum_noncortex_colors.append(colors[i])
        elif n in right_cerebrum_noncortex_list:
            right_cerebrum_noncortex_numbers.append(numbers[i])
            right_cerebrum_noncortex_names.append(names[i])
            right_cerebrum_noncortex_colors.append(colors[i])
        elif n in medial_cerebrum_noncortex_list:
            medial_cerebrum_noncortex_numbers.append(numbers[i])
            medial_cerebrum_noncortex_names.append(names[i])
            medial_cerebrum_noncortex_colors.append(colors[i])
        elif n in left_ventricle_list:
            left_ventricle_numbers.append(numbers[i])
            left_ventricle_names.append(names[i])
            left_ventricle_colors.append(colors[i])
        elif n in right_ventricle_list:
            right_ventricle_numbers.append(numbers[i])
            right_ventricle_names.append(names[i])
            right_ventricle_colors.append(colors[i])
        elif n in medial_ventricle_list:
            medial_ventricle_numbers.append(numbers[i])
            medial_ventricle_names.append(names[i])
            medial_ventricle_colors.append(colors[i])
        elif n in left_cerebellum_cortex_list:
            left_cerebellum_cortex_numbers.append(numbers[i])
            left_cerebellum_cortex_names.append(names[i])
            left_cerebellum_cortex_colors.append(colors[i])
        elif n in right_cerebellum_cortex_list:
            right_cerebellum_cortex_numbers.append(numbers[i])
            right_cerebellum_cortex_names.append(names[i])
            right_cerebellum_cortex_colors.append(colors[i])
        elif n in left_cerebellum_noncortex_list:
            left_cerebellum_noncortex_numbers.append(numbers[i])
            left_cerebellum_noncortex_names.append(names[i])
            left_cerebellum_noncortex_colors.append(colors[i])
        elif n in right_cerebellum_noncortex_list:
            right_cerebellum_noncortex_numbers.append(numbers[i])
            right_cerebellum_noncortex_names.append(names[i])
            right_cerebellum_noncortex_colors.append(colors[i])
        elif n in medial_cerebellum_noncortex_list:
            medial_cerebellum_noncortex_numbers.append(numbers[i])
            medial_cerebellum_noncortex_names.append(names[i])
            medial_cerebellum_noncortex_colors.append(colors[i])
        elif n in brainstem_list:
            brainstem_numbers.append(numbers[i])
            brainstem_names.append(names[i])
            brainstem_colors.append(colors[i])
        elif n in extra_list:
            extra_numbers.append(numbers[i])
            extra_names.append(names[i])
            extra_colors.append(colors[i])
        else:
            misc_numbers.append(numbers[i])
            misc_names.append(names[i])
            misc_colors.append(colors[i])

    # ------------------------------------------------------------------------
    # Aggregate lists of numbers, names, and colors:
    # ------------------------------------------------------------------------
    ventricle_numbers = left_ventricle_numbers + right_ventricle_numbers + \
                        medial_ventricle_numbers
    ventricle_names = left_ventricle_names + right_ventricle_names + \
                      medial_ventricle_names
    ventricle_colors = left_ventricle_colors + right_ventricle_colors + \
                       medial_ventricle_colors

    cerebrum_cortex_DKT31_numbers = left_cerebrum_cortex_DKT31_numbers + \
                                    right_cerebrum_cortex_DKT31_numbers
    cerebrum_cortex_DKT31_names = left_cerebrum_cortex_DKT31_names + \
                                  right_cerebrum_cortex_DKT31_names
    cerebrum_cortex_DKT31_colors = left_cerebrum_cortex_DKT31_colors + \
                                   right_cerebrum_cortex_DKT31_colors
    cerebrum_cortex_numbers = left_cerebrum_cortex_numbers + \
                              right_cerebrum_cortex_numbers
    cerebrum_cortex_names = left_cerebrum_cortex_names + \
                            right_cerebrum_cortex_names
    cerebrum_cortex_colors = left_cerebrum_cortex_colors + \
                             right_cerebrum_cortex_colors
    cerebrum_noncortex_numbers = left_cerebrum_noncortex_numbers + \
                                 right_cerebrum_noncortex_numbers + \
                                 medial_cerebrum_noncortex_numbers + \
                                 ventricle_numbers
    cerebrum_noncortex_names = left_cerebrum_noncortex_names + \
                               right_cerebrum_noncortex_names + \
                               medial_cerebrum_noncortex_names + \
                               ventricle_names
    cerebrum_noncortex_colors = left_cerebrum_noncortex_colors + \
                                right_cerebrum_noncortex_colors + \
                                medial_cerebrum_noncortex_colors + \
                                ventricle_colors
    left_cerebrum_numbers = left_cerebrum_cortex_numbers + \
                            left_cerebrum_noncortex_numbers
    left_cerebrum_names = left_cerebrum_cortex_names + \
                          left_cerebrum_noncortex_names
    left_cerebrum_colors = left_cerebrum_cortex_colors + \
                           left_cerebrum_noncortex_colors
    right_cerebrum_numbers = right_cerebrum_cortex_numbers + \
                             right_cerebrum_noncortex_numbers
    right_cerebrum_names = right_cerebrum_cortex_names + \
                           right_cerebrum_noncortex_names
    right_cerebrum_colors = right_cerebrum_cortex_colors + \
                            right_cerebrum_noncortex_colors
    cerebrum_numbers = left_cerebrum_numbers + right_cerebrum_numbers
    cerebrum_names = left_cerebrum_names + right_cerebrum_names
    cerebrum_colors = left_cerebrum_colors + right_cerebrum_colors

    left_cerebellum_numbers = left_cerebellum_cortex_numbers + \
                              left_cerebellum_noncortex_numbers
    left_cerebellum_names = left_cerebellum_cortex_names + \
                            left_cerebellum_noncortex_names
    right_cerebellum_numbers = right_cerebellum_cortex_numbers + \
                               right_cerebellum_noncortex_numbers
    right_cerebellum_names = right_cerebellum_cortex_names + \
                             right_cerebellum_noncortex_names

    cerebellum_cortex_numbers = left_cerebellum_cortex_numbers + \
                                right_cerebellum_cortex_numbers
    cerebellum_cortex_names = left_cerebellum_cortex_names + \
                              right_cerebellum_cortex_names
    cerebellum_cortex_colors = left_cerebellum_cortex_colors + \
                               right_cerebellum_cortex_colors
    cerebellum_noncortex_numbers = left_cerebellum_noncortex_numbers + \
                                   right_cerebellum_noncortex_numbers + \
                                   medial_cerebellum_noncortex_numbers
    cerebellum_noncortex_names = left_cerebellum_noncortex_names + \
                                 right_cerebellum_noncortex_names + \
                                 medial_cerebellum_noncortex_names
    cerebellum_noncortex_colors = left_cerebellum_noncortex_colors + \
                                  right_cerebellum_noncortex_colors
    cerebellum_numbers = cerebellum_cortex_numbers + \
                         cerebellum_noncortex_numbers
    cerebellum_names = cerebellum_cortex_names + cerebellum_noncortex_names
    cerebellum_colors = cerebellum_cortex_colors + cerebellum_noncortex_colors

    label_numbers = cerebrum_numbers + cerebellum_numbers + \
                    brainstem_numbers + extra_numbers
    label_names = cerebrum_names + cerebellum_names + \
                  brainstem_names + extra_names
    label_colors = cerebrum_colors + cerebellum_colors + \
                   brainstem_colors + extra_colors

    # ------------------------------------------------------------------------
    # Colormap:
    # ------------------------------------------------------------------------
    colormap = []
    colormap_normalized = []
    for i, x in enumerate(label_colors):
        colormap.append([label_numbers[i], 1, x[0], x[1], x[2]])
        colormap_normalized.append([label_numbers[i], 1,
                                    x[0]/255.0, x[1]/255.0, x[2]/255.0])

    # ------------------------------------------------------------------------
    # Sulcus names from the DKT labeling protocol:
    # ------------------------------------------------------------------------
    sulcus_names = [
        "frontomarginal sulcus",
        "superior frontal sulcus",
        "inferior frontal sulcus",
        "precentral sulcus",
        "central sulcus",
        "postcentral sulcus",
        "intraparietal sulcus",
        "primary intermediate sulcus/1st segment of post. sup. temporal sulcus",
        "sylvian fissure",
        "lateral occipital sulcus",
        "anterior occipital sulcus",
        "superior temporal sulcus",
        "inferior temporal sulcus",
        "circular sulcus",
        "1st transverse temporal sulcus and Heschl sulcus",
        "cingulate sulcus",
        "paracentral sulcus",
        "parietooccipital fissure",
        "calcarine fissure",
        "superior rostral sulcus",
        # "callosal sulcus",  # REMOVED June 2016: requires corpus callosum
        "lateral H-shaped orbital sulcus",
        "olfactory sulcus",
        "occipitotemporal sulcus",
        "collateral sulcus"]
    sulcus_numbers = list(range(len(sulcus_names)))

    sulcus_names_abbr = [
        "fms",
        "sfrs",
        "ifrs",
        "prcs",
        "cs",
        "pocs",
        "itps",
        "pis/csts1",
        "ls",
        "locs",
        "aocs",
        "sts",
        "its",
        "crs",
        "ftts/hs",
        "cgs",
        "pcs",
        "pos",
        "ccs",
        "sros",
        # "cas",  # REMOVED June 2016: requires corpus callosum
        "lhos",
        "olfs",
        "ots",
        "cos"]

    # ------------------------------------------------------------------------
    # Lists of label pairs that define sulcus boundaries (or fundi)
    # according to the DKT labeling protocol.
    # 1000 [lh] or 2000 [rh] are added to these numbers below.
    # ------------------------------------------------------------------------
    pair_lists = [
        [[12, 28]],
        [[3, 28], [27, 28]],
        [[3, 18], [3, 19], [3, 20], [18, 27], [19, 27], [20, 27]],
        [[24, 28], [3, 24], [24, 27], [18, 24], [19, 24], [20, 24]],
        [[22, 24]],
        [[22, 29], [22, 31]],
        [[29, 31], [8, 29]],
        [[8, 31]],
        [[30, 31]],
        [[8, 11], [11, 29]],
        [[11, 15], [9, 11]],
        [[15, 30]],
        [[9, 15]],
        [[12, 35], [30, 35], [34, 35], [2, 35], [10, 35], [23, 35], [26, 35],
         [22, 35], [24, 35], [31, 35]],
        [[30, 34]],
        [[2, 14], [10, 14], [14, 23], [14, 26], [2, 28], [10, 28], [23, 28],
         [26, 28],
         [2, 17], [10, 17], [17, 23], [17, 26], [17, 25]],
        [[17, 28]],
        [[5, 25]],
        [[13, 25], [2, 13], [10, 13], [13, 23], [13, 26]],
        [[14, 28]],
        # [[2, 4], [4, 10], [4, 23], [4, 26]],  # REMOVED June 2016
        [[3, 12], [12, 27], [12, 18], [12, 19], [12, 20]],
        [[12, 14]],
        [[7, 9], [7, 11]],
        [[6, 7], [7, 16], [7, 13]]]

    relabel = True

    left_sulcus_label_pair_lists = []
    right_sulcus_label_pair_lists = []
    unique_sulcus_label_pairs = []  # unique sorted label pairs
    for pair_list in pair_lists:
        left_pairs = []
        right_pairs = []
        for pair in pair_list:
            left_pair = [1000 + pair[0], 1000 + pair[1]]
            right_pair = [2000 + pair[0], 2000 + pair[1]]
            left_pairs.append(left_pair)
            right_pairs.append(right_pair)
            if relabel:
                if left_pair not in unique_sulcus_label_pairs:
                    unique_sulcus_label_pairs.append(left_pair)
                if right_pair not in unique_sulcus_label_pairs:
                    unique_sulcus_label_pairs.append(right_pair)
            else:
                if pair not in unique_sulcus_label_pairs:
                    unique_sulcus_label_pairs.append(pair)
        left_sulcus_label_pair_lists.append(left_pairs)
        right_sulcus_label_pair_lists.append(right_pairs)
    if relabel:
        sulcus_label_pair_lists = left_sulcus_label_pair_lists + \
                                  right_sulcus_label_pair_lists
    else:
        sulcus_label_pair_lists = pair_lists


"""
# ------------------------------------------------------------------------
# Variant of the DKT cerebral cortical labeling protocol with 25 labels:
# ------------------------------------------------------------------------
# Region numbers:
# DKT31 to DKT25: [[10,23,26,27,19,20], [2,2,2,3,18,18]]
left_cerebrum_cortex_numbers_DKT25 = list(range(1002, 1036))
for n in [1004, 1010, 1019, 1020, 1023, 1026, 1027, 1032, 1033]:
    left_cerebrum_cortex_numbers_DKT25.remove(n)
right_cerebrum_cortex_numbers_DKT25 = list(range(2002, 2036))
for n in [2004, 2010, 2019, 2020, 2023, 2026, 2027, 2032, 2033]:
    right_cerebrum_cortex_numbers_DKT25.remove(n)
cerebrum_cortex_numbers_DKT25 = left_cerebrum_cortex_numbers_DKT25 + \
                                right_cerebrum_cortex_numbers_DKT25
# Consolidate region labels:
left_cerebrum_cortex_names_DKT25 = []
for ilabel, label_number in enumerate(left_cerebrum_cortex_numbers):
    if label_number in left_cerebrum_cortex_numbers_DKT25:
        if label_number == 1002:
            left_cerebrum_cortex_names_DKT25.append('left cingulate')
        elif label_number == 1003:
            left_cerebrum_cortex_names_DKT25.append('left middle frontal')
        elif label_number == 1018:
            left_cerebrum_cortex_names_DKT25.append('left inferior frontal')
        else:
            left_cerebrum_cortex_names_DKT25. \
                append(left_cerebrum_cortex_names[ilabel])
right_cerebrum_cortex_names_DKT25 = []
for ilabel, label_number in enumerate(right_cerebrum_cortex_numbers):
    if label_number in right_cerebrum_cortex_numbers_DKT25:
        if label_number == 2002:
            right_cerebrum_cortex_names_DKT25.append('right cingulate')
        elif label_number == 2003:
            right_cerebrum_cortex_names_DKT25.append('right middle frontal')
        elif label_number == 2018:
            right_cerebrum_cortex_names_DKT25.append('right inferior frontal')
        else:
            right_cerebrum_cortex_names_DKT25. \
                append(right_cerebrum_cortex_names[ilabel])
cerebrum_cortex_names_DKT25 = \
    left_cerebrum_cortex_names_DKT25 + right_cerebrum_cortex_names_DKT25
"""
"""
# 2-D cortical map (UNDER CONSTRUCTION):
#
# Each gyrus has one or more sulcus boundaries -- rough breakdown:
#
# labels:  28  3   18  30  15  9   7   16   17    12   14    29  8      24   22      5    25   6   21    2    35   31
#
# Gyrii:  sFG mFG iFG sTG mTG iTG Fus pHip pCen  lOrb mOrb  sPL iPL    prCG poCG    Cun prCun ent pCalc Cing Ins sMarg
#         ---|---|---|---|---|---|---|----|----  ---- ----  ---|---  -|----|----|-  ---|----- --- ----- ---- --- -----
# Sulci:    sFS iFS Syl sTS iTS TO? Coll?        OrbF OrbT    iPS    prCS  CS poCS                 Cing Ins
#
# indices:   1   2   3   4   5   6   7      8    9      10      11   12  13   14  15  16   17  18
"""
